/*
 * Author: Milind Changire
 *
 * Copyright Corel Comptuer Corp (c) 1997-1999
 *
 * Video Capture driver for the Philips SAA7111 interfacing via reg 3CF/B6 
 * of the CyberPro2000 using the I2C interface.
 * The IGS SDK for video capture has been used to develop this device driver.
 * http://www.igst.com
 *
 * Note:
 *   Use tabstop=2 for viewing/editing.
 *
 * Note:
 *   All operations handle YUV422 data and NOT RGB data.
 *   The YUV to RGB conversion routines can be used to convert a captured frame
 *   to create a photograph or portrait for a Video Conferencing application.
 *
 */

#include <linux/config.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <sys/mman.h>
#include <asm/segment.h>

#include "igs.h"				/* IGS register list and function declarations */
#include "video.h"			/* The current states of the driver */

#define VIDEO_MAJOR 81	/* Major number of the /dev/video device */

/***************************************************************************/

static int		video_open(struct inode* inodep, struct file* filep);

static void		video_release(struct inode* inodep, struct file* filep);

static int		video_ioctl(struct inode* inodep,
													struct file* filep, 
													unsigned int cmd,
													unsigned long arg);

static int 		video_read(	struct inode* inodep,
													struct file* filep,
													char* pchBufferPointer,
													int nBufferSize);

static int 		video_write(struct inode* inodep,
													struct file* filep,
													const char* pchBufferPointer,
													int nBufferSize);

static struct file_operations video_fops = {
	NULL,							/* no special lseek              */
	video_read,
	video_write,
	NULL,							/* no special readdir            */
	NULL,							/* no special select             */
	video_ioctl,
	NULL,							/* no special mmap               */
	video_open,
	video_release,
	NULL,							/* no special fsync              */
	NULL,							/* no special fasync             */
	NULL,							/* no special check_media_change */
	NULL							/* no special revaldate          */
};

static int setup_video(void);
static int setup_vmem(void);
static void enable7111(int state);
static void open_init(void);
extern int set_CPLD(int bitset, int bitmask);

/* 	The memory size that this driver will need. 
		(Defaulted to about 2xCIF images) 
		Capture requires slightly larger than CIF to capture 
		capture CIF image due to garbage lines. 
*/
static int		DRIVER_MEMORY_SIZE				= 352 * 288 * 2 + 352 * 294 * 2;
static const	char video_device_name[]	= "video";
static int		video_major								= 0;

/*SDOC***********************************************************************

	Name:			enable7111

	Action:		enable/disable SAA7111 decoder (off to reduce power consumption)

	Params:		None	

	Returns:	None

	Comments:	we want to write a bit pattern 0X10 to CPLD, where X - 7111 state
						1 = disabled, 0 = enabled

***********************************************************************EDOC*/
static void enable7111(int state)
{
	int i=0;

	set_CPLD(state<<2, 0x04);

	if (!state)	/* when we turn on 7111, need to introduce some delay */
	{
		for (i=0;i<1000;i++) 
			udelay(1000);
	}
}

/*SDOC***********************************************************************

	Name:			setup_video

	Action:		Initializes IGS CyberPro2000 registers required for video capture.
						Actual values for displaying the captured video are set in 
						video_ioctl.

	Params:		None	

	Returns:	0				- ok
						-ENXIO	- if current video mode does not support video capture.

	Comments:

***********************************************************************EDOC*/
static int setup_video()
{
	int nDecoder = 0;
	int nReturnValue = -1;

	/* Device has to be used within graphics mode (svga or X) */
	if (FALSE == IGS_IsGraphics())
	{
		printk("video.c: Video capture not supported in text mode.\n");
		return -EIO;
	}

	enable7111(0);	/* turn on 7111 */
	VDev.gn7111On = 1;

#ifdef DEBUG
printk("video.c: Enabling SAA7111 decoder.\n");
#endif 

	if (NODECODER == (nDecoder = DetectDecoder()))
	{
		printk("video.c: No decoder found.\n");
	}
	else if ((PHILIPS7111 == nDecoder) && (TRUE == InitPhil7111()))
	{
		VDev.gScreen.x_ext = GetCurrentScreenSize(X_SCREEN);	/* 1 based value */
		VDev.gScreen.y_ext = GetCurrentScreenSize(Y_SCREEN);  /* 1 based value */
		VDev.gScreen.width = VDev.gScreen.x_ext;
		VDev.gScreen.height = VDev.gScreen.y_ext;

		Out_Video_Reg_M(0x33, 0x0c, 0xf3);	/* Enabling coprocessor */

		VDev.gScreen.bytes_per_pixel = GetBpp() >> 3;	/* 1-8bpp, 2-16bpp, 3-24bpp */

#ifdef DEBUG
printk("video.c: Bytes per pixel: %d\n", VDev.gScreen.bytes_per_pixel);
#endif

		/* Initialize graphics engine register */
		Set_IGS_Px_Map(VDev.gScreen.x_ext-1, VDev.gScreen.bytes_per_pixel-1);

  	InitXlatTab();   /* first YUV->RGB xlat table */

		/* Initialize video parameters */
		if (setup_vmem() && GetVideoFIFO())
		{
			/* Initialize IGS registers */
			IGS_InitVideo();

#if SUPPORTVUNDERSCAN
			/* get underscan information */
			VDev.wV_Underscan = (ReadReg(0x3d4,0x45) & 0x10) >>4;
			if (VDev.wV_Underscan==1)
			{
				VDev.iMaxLines  = ReadReg(0x3d4,0x44) + 0x11;
				VDev.iWhichLine = VDev.iMaxLines >> 4;
				VDev.iMaxLines  = VDev.iMaxLines & 0x0F;
			}
#endif

			nReturnValue = 0;
		
			VDev.gnVideoInitialized = 1;		/* video device has been initialized */
		}
		else
		{
			printk("video.c: Error setting up vmem and VideoFIFO.\n");
			nReturnValue = -ENXIO;
		}
	}
	else
  {
    printk("video.c: Unable to initialize Philips SAA 7111 decoder.\n");
  }

	return (nReturnValue);
}

/*SDOC***********************************************************************

	Name:			setup_vmem	

	Action:		Setup video memory management. Make sure we have enough to do
						minimum capture/display requirement.

	Params:		None	

	Returns:	FALSE - if not enough memory or does not support capture mode
						TRUE  - ok

	Comments:

***********************************************************************EDOC*/
static int setup_vmem()
{
	int nReturnValue = FALSE;
	int i = 0;

#ifdef DEBUG
printk("video.c: Current screen resolution: %dx%dx%d\n", VDev.gScreen.x_ext,
				VDev.gScreen.y_ext, VDev.gScreen.bytes_per_pixel);
#endif

	if (XVideoBuffer.gnWithinX)
	{
#ifdef DEBUG
printk("video.c: Apps running INSIDE X window.\n");
#endif
		VDev.gAMI.gunTotalAvailableMemory = XVideoBuffer.vbuf.height * 
	 																			XVideoBuffer.vbuf.width;
		VDev.gAMI.gulAvailableMemoryStart = *((int*)XVideoBuffer.vbuf.base);
		VDev.gAMI.gulCaptureStart = VDev.gAMI.gulAvailableMemoryStart;
		VDev.gAMI.gulCaptureEnd = VDev.gAMI.gulCaptureStart;

		/* Capture Ends varies with different capture resolution and will
				be set again within VIDIOCSWIN function. */

#ifdef DEBUG
printk("video.c: Total usable memory is: %d\n",
				VDev.gAMI.gunTotalAvailableMemory);
printk("video.c: Available memory start: %ld\n",
				VDev.gAMI.gulAvailableMemoryStart);
printk("video.c: Capture Start: %ld\n", VDev.gAMI.gulCaptureStart);
#endif

		/* Do not specify size for the display regions. It will be initialized
				when they are actually needed in VIDIOC_COREL_SDWIN */
		for (i=0; i<MAX_NUM_OF_DISPLAY_REGION; i++)
		{
			VDev.gAMI.gulDisplayStart[i]  = 0;
			VDev.gAMI.gulRegionSize[i]    = 0;
		}

		nReturnValue = TRUE;
	}
	else
	{	/* 
			Apps are opening the device outside X environment.
			The driver will assume the control of all the free available
			memory in the video card.
		*/
#ifdef DEBUG
printk("video.c: Apps running OUTSIDE X window.\n");
#endif

		VDev.gScreen.display_memory_used	=	VDev.gScreen.x_ext *
																				VDev.gScreen.y_ext *
																				VDev.gScreen.bytes_per_pixel;
		VDev.gScreen.total_video_memory		= GetVMemSize() * 2 * 1024 * 1024;
	
		VDev.gAMI.gunTotalAvailableMemory = VDev.gScreen.total_video_memory -
																				VDev.gScreen.display_memory_used -
																				BUFFER_REGION_SIZE;
#ifdef DEBUG
printk("video.c: Total usable memory is: %d\n", 
				VDev.gAMI.gunTotalAvailableMemory);
#endif

		/* check if available memory size meets minimum requirement */
		/* Minimum memory requirement is enough to do a CIF capture and overlay
				a CIF video pumped through network */
		if (DRIVER_MEMORY_SIZE <= VDev.gAMI.gunTotalAvailableMemory)
		{
			VDev.gAMI.gulAvailableMemoryStart = VDev.gScreen.display_memory_used +
																					BUFFER_REGION_SIZE;
			VDev.gAMI.gulCaptureStart = VDev.gAMI.gulAvailableMemoryStart;
			VDev.gAMI.gulCaptureEnd = VDev.gAMI.gulCaptureStart;

#ifdef DEBUG
printk("video.c: Available memory start: %ld\n", 
				VDev.gAMI.gulAvailableMemoryStart);
printk("video.c: Capture Start: %ld\n", VDev.gAMI.gulCaptureStart);
#endif

			/* Capture Ends varies with different capture resolution and will
					be set again within VIDIOCSWIN function. */ 

			/* Do not specify size for the display regions. It will be initialized
					when they are actually needed in VIDIOC_COREL_SDWIN */ 
			for (i=0; i<MAX_NUM_OF_DISPLAY_REGION; i++)
			{
				VDev.gAMI.gulDisplayStart[i]	= 0;
				VDev.gAMI.gulRegionSize[i]		= 0;
			}

			nReturnValue = TRUE;
		}
		else
		{
			printk("video.c: Not enough video memory for driver.\n"); 
		}
	}

	return (nReturnValue);
}

/*SDOC***********************************************************************

	Name:			open_init

	Action:		initialize the driver to default values within "open()".

	Params:		none

	Returns:	none

	Comments:

***********************************************************************EDOC*/
static void open_init(void)
{
	/* Initializing video capabilities */
	strcpy(VDev.gVideoCapability.name, "NetWinder Video Driver");
	VDev.gVideoCapability.type			=	VID_TYPE_CAPTURE |
																		VID_TYPE_CHROMAKEY |
																		VID_TYPE_SCALES  |
																		VID_TYPE_SUBCAPTURE;
	VDev.gVideoCapability.channels	= 1;
	VDev.gVideoCapability.audios		= 0;
	/* maxwidth and maxheight have been defined */
	VDev.gVideoCapability.minwidth	= MIN_CAPTURE_WIDTH;
	VDev.gVideoCapability.minheight = MIN_CAPTURE_HEIGHT;

	/* Initializing video picture parameters */ 
	VDev.gVideoPicture.brightness		= DEFAULT_CAPTURE_BRIGHTNESS;
	VDev.gVideoPicture.hue					= DEFAULT_CAPTURE_HUE;
	VDev.gVideoPicture.colour				= DEFAULT_CAPTURE_COLOUR; 
	VDev.gVideoPicture.contrast			= DEFAULT_CAPTURE_CONTRAST;
	VDev.gVideoPicture.whiteness		= DEFAULT_CAPTURE_WHITENESS;
	VDev.gVideoPicture.depth				= DEFAULT_CAPTURE_DEPTH;
	VDev.gVideoPicture.palette			= DEFAULT_CAPTURE_PALETTE;

	/* Initializing all video channels parameters */
	VDev.gnSelectedChannel					= 1;	/* defaulted to use channel 1 */
	VDev.gVideoChannel[0].channel 	= 1;
	strcpy(VDev.gVideoChannel[0].name, "Video camera");
	VDev.gVideoChannel[0].tuners		= 0;
	VDev.gVideoChannel[0].flags			= 0;
	VDev.gVideoChannel[0].type			= VIDEO_TYPE_CAMERA;
	VDev.gVideoChannel[0].norm			= 0;

	/* Initialize all video units parameters */
	VDev.gVideoUnit.video						= 1;	/* others are zero */

	/* No display window is associated with any particular display region yet. */
	VDev.gDWinRegion[0]							= -1;	/* Main video */
	VDev.gDWinRegion[1]							= -1;	/* V2 video */
	VDev.gDWinRegion[2]							= -1; /* X2 video */
						
	VDev.gIconBase									= -1;	/* No video window is associated
																						with icon video yet */
	VDev.gShiftDest									= -1;	/* No video window is associated
																						with shifted video yet */
	VDev.gnReadOn										= 1;	/* Allow read() by default */

#if SUPPORTVUNDERSCAN
	VDev.wV_Underscan	= 0;
	/* # of scan lines within which we drop one scan line */
	VDev.iMaxLines		= 0;	
	VDev.iWhichLine		= 0;	/* which scan line we drop */
#endif
}

/*SDOC***********************************************************************

	Name:			video_open

	Action:		open video device to perform capture/overlay

	Params:		struct file* - file descriptor

	Returns:	-1 - if fail
						file descriptor - if successful

	Comments:	Device driver exported function
						/dev/video0 has to be opened before /dev/video1 or /dev/video2.	
						/dev/video0 can be opened more than once. 
						/dev/video1 and /dev/video2 can ONLY be opened once.

***********************************************************************EDOC*/
static int video_open(struct inode* inodep, struct file* filep)
{
	int nReturnValue	= -1;
	int nMinor				= MINOR(inodep->i_rdev); 

	switch (nMinor)
	{
		case 0:	/* Minor number = 0, Controlling Capture and Main video window */
		{
			if (0 == VDev.gnOpenCount)
			{
				memset(&VDev, 0, sizeof(VDev));		/* clear the whole device state */
				open_init();
			}
			nReturnValue = 0;
			VDev.gnMainOpen++;
			VDev.gnOpenCount++;
			MOD_INC_USE_COUNT;
		}
		break;

		case 1:	/* opening V2 video window */
		{
			/* /dev/video0 has to be opened first */
			/* /dev/video1 can only be opened once */
			/* We don't need to reserve memory space for opening V2 and X2 video
					window because we won't need any memory if we only overlay
					local capture image onto those overlay windows
			*/
			if ((VDev.gnMainOpen>0) && (!VDev.gnV2Open))
			{
				VDev.gnV2Open = 1;
       	VDev.gnOpenCount++;
       	MOD_INC_USE_COUNT;
				nReturnValue = 0;
			}
			else
			{
				printk("video.c: Error opening /dev/video1.\n");
			}
		}
		break;

		case 2:	/* opening X2 video window */
		{
			if ((VDev.gnMainOpen>0) && (!VDev.gnX2Open))
			{
				VDev.gnX2Open = 1;
       	VDev.gnOpenCount++;
       	MOD_INC_USE_COUNT;
				nReturnValue = 0;
			}
			else
			{
				printk("video.c: Error opening /dev/video2.\n");
			}
		}
		break;

		default:
			printk("video.c: Invalid minor number.\n");

	}/* end-switch */

	return (nReturnValue);
}

/*SDOC***********************************************************************

	Name:			video_release

	Action:		close video device

	Params:		struct file* - file descriptor

	Returns:	None

	Comments:	Device driver exported function
						When /dev/video0 is being closed, any opened /dev/video1 or 
						/dev/video2 will also be closed as well.

***********************************************************************EDOC*/
static void video_release(struct inode* inodep, struct file* filep)
{
	switch(MINOR(inodep->i_rdev))	
	{
		case 0:
		{
			if (VDev.gnV2Open)
			{
				IGS_OverlayOff_V2();
				IGS_VideoCleanUp_V2();
				VDev.gnV2Open = 0;
				VDev.gnOpenCount--;
				MOD_DEC_USE_COUNT;
			}

			if (VDev.gnX2Open) 
			{
				IGS_OverlayOff_X2();
				IGS_VideoCleanUp_X2();
				VDev.gnX2Open = 0;
				VDev.gnOpenCount--;
				MOD_DEC_USE_COUNT;
			}

			if (VDev.gnMainOpen > 0)
			{
				IGS_VideoOff();
				IGS_VideoCleanUp();
				VDev.gnMainOpen--;
				VDev.gnOpenCount--;
				if (0 == VDev.gnOpenCount)
					VDev.gnVideoInitialized = 0;
				MOD_DEC_USE_COUNT;
			}
		}
		break;
	
		case 1:
		{
			if (VDev.gnV2Open)
			{
				IGS_OverlayOff_V2();
				IGS_VideoCleanUp_V2();
				VDev.gnV2Open = 0;
				VDev.gnOpenCount--;
				MOD_DEC_USE_COUNT;
			}
		}
		break;
		
		case 2:
		{
			if (VDev.gnX2Open) 
			{
				IGS_OverlayOff_X2();
				IGS_VideoCleanUp_X2();
				VDev.gnX2Open = 0;
				VDev.gnOpenCount--;
				MOD_DEC_USE_COUNT;
			}
		}
		break;
		
		default:
		{
			printk("video.c: Invalid minor number.\n");
		}
	}

	if ((0 == VDev.gnOpenCount) && VDev.gn7111On)
	{
#ifdef DEBUG
printk("video.c: Disabling SAA7111 decoder.\n");
#endif
		enable7111(1);	/* disable 7111 */
	}
}

/*SDOC***********************************************************************

	Name:			video_read

	Action:		retrieve a captured frame

	Params:		struct file* - file descriptor
						char* - pointer to the user space buffer.
						int   - size of the user space buffer.

	Returns:	Number of bytes requested - if successful
						-1 - if fail

	Comments:	

***********************************************************************EDOC*/
static int video_read(struct inode* inodep,
											struct file* filep,
											char* pchBufferPointer,
											int nBufferSize)
{
	unsigned char* puchDataPtr	= (void*)0;
#if 0
	long long current_ptr				= filep->f_pos;
#endif
	int nDataSize								= 0;

	/* precaution taken to avoid problem during video mode switching */
	if (!VDev.gnReadOn)
		return -1;

	/* find the pointer to first byte of good capture data */
	puchDataPtr = (unsigned char*)
								(	(long)((0xd0000000L + VDev.gAMI.gulCaptureStart) +
									(VDev.gnGarbageLines * VDev.gnActualCaptureWidth * 2)));

#if 0
printk("puchDataPtr = %p\n", puchDataPtr);
filep->f_pos = (long long)puchDataPtr+50688;
printk("filep->f_pos = %p\n", (unsigned char*)filep->f_pos);
#endif

#ifdef EXPERIMENT 
	/* opened using "cat" command... */ 
	if (!VDev.gnVideoInitialized)
	{
		printk("Opening using \"cat\" command...\n");
		if (0 != setup_video())
		{
			printk("video.c: Error setting up video device.\n");
			return -1;
		}
		VDev.gnVideoInitialized = 0;
		memcpy_tofs(pchBufferPointer, puchDataPtr, 176*144*2);
		filep->f_pos += 176*144*2;
		return (176*144*2);
	}
#endif

	switch (VDev.gVideoPicture.palette)
	{
		default:
		case VIDEO_PALETTE_YUYV:
		{
			nDataSize = VDev.gnActualCaptureWidth * VDev.gnActualCaptureHeight * 2;
			if (nBufferSize == nDataSize)
			{
				memcpy_tofs(pchBufferPointer, puchDataPtr, nDataSize);
				filep->f_pos = 0;	/* reset the position pointer */
			}
			else
			{
				nDataSize = -1;
			}
		}
		break;
		
		case VIDEO_PALETTE_RGB565:
		{
			nDataSize = VDev.gnActualCaptureWidth * VDev.gnActualCaptureHeight * 2;
			if (nBufferSize == nDataSize)
			{ 
				ConvertYUVtoRGB16(pchBufferPointer, puchDataPtr, 
													VDev.gnActualCaptureWidth, 
													VDev.gnActualCaptureHeight);
			}
			else
			{
				nDataSize = -1;
			}
		}
		break;
			
		case VIDEO_PALETTE_RGB24:
		{
			nDataSize = VDev.gnActualCaptureWidth * VDev.gnActualCaptureHeight * 3;
			if (nBufferSize == nDataSize)
			{
				ConvertYUVtoRGB24(pchBufferPointer, puchDataPtr, 
													VDev.gnActualCaptureWidth, 
													VDev.gnActualCaptureHeight);
			}
			else
			{
				nDataSize = -1;
			}
		}
		break;
	}
				
	return(nDataSize);
}

/*SDOC***********************************************************************

	Name:			video_write

	Action:		put a frame onto the screen

	Params:		struct file* - file descriptor
						char* - pointer to the user space buffer.
						int   - size of the user space buffer.

	Returns:	Number of bytes written - if successful
						-1 - if fail

	Comments:	Device driver exported function

***********************************************************************EDOC*/
static int video_write(	struct inode* inodep,
												struct file* filep,
												const char* pchBufferPointer,
												int nBufferSize)
{
	int nMinor = MINOR(inodep->i_rdev);
	unsigned char* puchDataPtr = (void*)0;
	int nReturnValue = 0;

	if (-EFAULT == verify_area(VERIFY_READ, pchBufferPointer, nBufferSize))
	{
		printk("video.c: write() caused -EFAULT on verify_area.\n");
		nReturnValue = -EFAULT;
	}
	else
	{
		/* 	If...
				Scale down,
				Displaying background for second type of PIP or
				Display freeze,
				video data are first put into TEMP region 
		*/
		if	(	(2 <= VDev.gnScaleDownFactor[nMinor]) || 
					(	VDev.gnPIPFlag[nMinor] && 
						VDev.gPIP[nMinor].type && 
						!VDev.gPIP[nMinor].order
					) || (VDev.gFreezeFlag[nMinor])
				)
		{
			/* find the first byte of the TEMP region */
			puchDataPtr = (unsigned char*)
										((long)(0xd0000000L + VDev.gAMI.gulTempStart));
		}
		else
		{
			/* find the first byte of the display region */
			puchDataPtr = (unsigned char*)
										(	(long)(0xd0000000L +
											VDev.gAMI.gulDisplayStart[(VDev.gDWinRegion[nMinor])])
										);
		}

		/* check if there is really enough space */
		if (nBufferSize <= VDev.gAMI.gulRegionSize[(VDev.gDWinRegion[nMinor])])
		{
			memcpy_fromfs(puchDataPtr, pchBufferPointer, nBufferSize);
			nReturnValue = nBufferSize;

			/* Using scale down? */
			if (2 <= VDev.gnScaleDownFactor[nMinor])
			{
				Corel_ScaleDown(nMinor);
			}

			/* Using PIP? */
			if (VDev.gnPIPFlag[nMinor])	
			{
				if (!VDev.gPIP[nMinor].type)	/* First type of PIP */
				{
					Corel_PIP1(nMinor);
				}
				else													/* Second type of PIP */
				{
					Corel_PIP2(nMinor, VDev.gPIP[nMinor].order);
				}
			}

			/* Using Icon? */
			if (VDev.gIconFlag)
			{
				if (VDev.gIcon.src == nMinor)		/* Is this video window the source? */
				{
					if (0 != VDev.gIconCurrent)
					{
						Corel_Icon(nMinor);
					}
					else
					{
#ifdef DEBUG
printk("video.c: Current Icon is 0.\n");
#endif
						nReturnValue = -1;
					}
				}
			}

			/* Using display freezing? */
			if (VDev.gFreezeFlag[nMinor])
			{
				Corel_DisplayFreeze(nMinor, VDev.gVideoFreeze[nMinor].type);
			}

			/* Using shift video? */
			if (VDev.gShiftFlag && (VDev.gVideoShift.src == nMinor))
			{
				Corel_ShiftVideo(nMinor);
			}
		}
		else	/* Not enough space */
		{
			nReturnValue = -1;
		}
	}/*end-else*/

	return (nReturnValue);
}

/*SDOC***********************************************************************

	Name:			video_ioctl

	Action:		control most of the capture/overlay functions and parameters

	Params:		struct file* - file descriptor
						unsigned int - the specific io control call name
						unsigned long - argument for the io control call

	Returns:	0  - if successful
						-1 - if fail

	Comments:	Device driver exported function

***********************************************************************EDOC*/
static int video_ioctl(	struct inode* inodep,
												struct file* filep,
												unsigned int cmd,
												unsigned long arg)
{
	int nReturnValue = 0;

	switch (cmd)
	{
    case VIDIOCGCAP:	/* Getting video device capabilities */
    {
			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_capability*)arg,
																	sizeof(VDev.gVideoCapability)))
      {
        printk("video.c: VIDIOCGCAP caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else 
			{
      	memcpy_tofs((struct video_capability*)arg, &(VDev.gVideoCapability), 
										sizeof(VDev.gVideoCapability));
			}
    }
    break;

		/* Called by X server to get the requested memory size */
		case VIDIOCGFBUF:
		{
			struct video_buffer VideoBuffer;
			
			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}
			
			VideoBuffer.base					= (void*)0;
			VideoBuffer.height				= 1;
			VideoBuffer.width					= DRIVER_MEMORY_SIZE;
			VideoBuffer.depth					= 0;
			VideoBuffer.bytesperline	= 0;
		
			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_buffer*)arg,
																	sizeof(VideoBuffer)))
      {
        printk("video.c: VIDIOCGFBUF caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else 
			{
				memcpy_tofs((struct video_buffer*)arg, &VideoBuffer,
										sizeof(VideoBuffer));
			}
		}
		break;

		/* 
			Called by the X server to set the available memory start pointer
			and available memory size 
		*/
		case VIDIOCSFBUF:
		{
			struct video_buffer VideoBuffer;

			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}
			
			if (-EFAULT == verify_area(	VERIFY_READ, (struct video_buffer*)arg,
																	sizeof(VideoBuffer)))
			{
				printk("video.c: VIDIOCSFBUF caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else
			{
				memcpy_fromfs(&VideoBuffer, (struct video_buffer*)arg,
											sizeof(VideoBuffer));

				/* Update the current state device structure */
				/* 
					When this function is called, it implies this driver will
					be used with X server 
				*/
				*((int*)XVideoBuffer.vbuf.base)	= *((int*)VideoBuffer.base);
				XVideoBuffer.vbuf.height				= VideoBuffer.height;
				XVideoBuffer.vbuf.width					= VideoBuffer.width;
				XVideoBuffer.vbuf.depth					= VideoBuffer.depth;
				XVideoBuffer.vbuf.bytesperline	= VideoBuffer.bytesperline;

#if 0
printk("Inside driver:\n");
printk("Memory start address: %d\n", *((int*)XVideoBuffer.vbuf.base));
printk("Memory size available: %d\n", XVideoBuffer.vbuf.height * 
																			XVideoBuffer.vbuf.width);
printk("Depth: %d\n", XVideoBuffer.vbuf.depth);
printk("Bytesperline: %d\n", XVideoBuffer.vbuf.bytesperline);
#endif

				/* 
					If the video device has already been opened, implies some app is using
					it outside X-window and the address provided the X should be
					ignored.  Garbage could be seen on X windows coz video driver
					and X server is fighting over physical video memory.
				*/
				XVideoBuffer.gnWithinX = !(VDev.gnVideoInitialized);
			}
		}
		break;

		case VIDIOCGWIN:		/* Getting capture window parameters */
		{
			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}
			
			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_window*)arg,
																	sizeof(VDev.gCaptureWindow)))
      {
        printk("video.c: VIDIOCGWIN caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else 
			{
				memcpy_tofs((struct video_window*)arg, &(VDev.gCaptureWindow),
										sizeof(VDev.gCaptureWindow));
			}
		}
		break;

		/* Should call Set image properties and set channel before this function
				if the desired image properties and channel are different */
		/* Note: only width and height parameters will be used in the passed-in 
			struct video_window */
		case VIDIOCSWIN:		/* Setting capture window parameters */
		{
			struct video_window		VideoWindow;
			unsigned long					ulOrigCaptureEnd = 0L;
			long int							lnCaptureSize = 0L;
			int										nRegion = 0;
			
			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}
			
			/* If no valid channel has been selected, return error */
			if (UNINITIALIZED == VDev.gnSelectedChannel)
			{
				nReturnValue = -1;
#ifdef DEBUG
printk("video.c: Channel selected is not valid.\n");
#endif
				break;
			}

			/* 
				NOTE:
				Ideally, bytes per pixel should vary with different capture format,
				but actually no matter which capture format has been chosen, byte per
				pixel is always 2 since only YUYV format data will always be
				written to the memory.
			*/ 

			if (-EFAULT == verify_area(	VERIFY_READ, (struct video_window*)arg,
																	sizeof(VideoWindow)))
			{
				printk("video.c: VIDIOCSWIN caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_fromfs(&VideoWindow, (struct video_window*)arg,
										sizeof(VideoWindow));				

			/* Validating the inputed values */
			VideoWindow.x										= -999;								/* Not used */
			VideoWindow.y										= -999;								/* Not used */
			VideoWindow.chromakey						= (__u32)0xFFFFFFFF;	/* Not used */
			VideoWindow.flags								= 0;									/* Not used */
			VideoWindow.clips								= (void*)0;						/* Not used */
			VideoWindow.clipcount						= 0;									/* Not used */

			/* Capture width has to be multiple of 4 and capture height is multiple
					of 2. */
			VideoWindow.width		= (VideoWindow.width  / 4) * 4;
			VideoWindow.height	= (VideoWindow.height / 2) * 2;

			if (VDev.gVideoCapability.minwidth > VideoWindow.width)
			{
				VideoWindow.width = VDev.gVideoCapability.minwidth;
			}
			if (VDev.gVideoCapability.minheight > VideoWindow.height)
			{
				VideoWindow.height = VDev.gVideoCapability.minheight;
			}
			if (VDev.gVideoCapability.maxwidth < VideoWindow.width)
			{
				VideoWindow.width = VDev.gVideoCapability.maxwidth;
			}
			if (VDev.gVideoCapability.maxheight < VideoWindow.height)
			{
				VideoWindow.height = VDev.gVideoCapability.maxheight;
			}

			/* Capture width larger than 352 and VDev.gnFullScreenOk is 0 is
				 not allowed. */
			if ((VideoWindow.width > MAXMPEG1) && (0 == VDev.gnFullScreenOk))
			{
				/* This capture resolution is not supported in this video mode.
					 Try to use a lower display resolution. */
#ifdef DEBUG
printk("video.c: Requested capture resolution is not supported in current video mode.\n");
#endif
				nReturnValue = -1;
				break;
			}
			else
			{
				/* Video mode supports the current capture resolution but do we have 
					 enough memory for the capturing? */

				/* Save the original capture end pointer */
				ulOrigCaptureEnd = VDev.gAMI.gulCaptureEnd;
#ifdef DEBUG
printk("video.c: Original capture end pointer is: %ld\n", ulOrigCaptureEnd);
#endif

				/* Find out the memory size needed. */
				lnCaptureSize = VideoWindow.width * 
												(VideoWindow.height + NTSC_EXTRA_CAP_LINES) * 2;

				/* check if there's any display region got initialized yet */
				if (0 == VDev.gAMI.gulDisplayStart[0])
				{	/* No display region has been initialized */
#ifdef DEBUG
printk("video.c: No display region has been initialized.\n");
#endif
					/* check if we have enough mem to do capture */
					if (	(lnCaptureSize + VDev.gAMI.gulCaptureStart) >
								(VDev.gAMI.gulAvailableMemoryStart + 
								VDev.gAMI.gunTotalAvailableMemory)
							)
					{
						/* capture resolution is too high */
#ifdef DEBUG
printk("video.c: Capture resolution is too high, reduce it.\n");
#endif
						nReturnValue = -1;
						break;
					}
				}
				else	/* some display region has been initialized */
				{
#ifdef DEBUG
printk("video.c: Some display region has been initialized.\n");
#endif
					/* checking which is the latest initialized display region */
					for (nRegion = MAX_NUM_OF_DISPLAY_REGION-1; nRegion >=0; nRegion--)
					{
						if (0 != VDev.gAMI.gulDisplayStart[nRegion])
						{
#ifdef DEBUG
printk("video.c: Region %d is the last initialized display region.\n", nRegion);
#endif
							break;	
						}
					}
					
					/* check if we still have enough mem to do capture */
					if ((lnCaptureSize + VDev.gAMI.gulCaptureStart) > 
							VDev.gAMI.gulDisplayStart[nRegion])	
					{
						/* capture resolution is too high */
#ifdef DEBUG
printk("video.c: Capture resolution is too high, reduce it.\n");
#endif
						nReturnValue = -1;
						break;
					}
				}

				/* Update capture end pointer */
				VDev.gAMI.gulCaptureEnd = lnCaptureSize + VDev.gAMI.gulCaptureStart;
#ifdef DEBUG
printk("video.c: Capture End: %ld\n", VDev.gAMI.gulCaptureEnd);
#endif

				/* Is there any temp region? If so, need to adjust the temp region */
				if (0 != VDev.gAMI.gulTempStart)
				{
#ifdef DEBUG
printk("video.c: TEMP region has already been initialized.\n");
#endif
					if (-1 == Set_Temp_Region(VDev.gAMI.gulTempSize))
					{
#ifdef DEBUG
printk("video.c: After setting capture region, failed to adjust the temp region.\n");
#endif
						nReturnValue = -1;
						/* restore old capture end pointer value */
#ifdef DEBUG
printk("video.c: Restoring old capture end pointer to: %ld\n",ulOrigCaptureEnd);
#endif
						VDev.gAMI.gulCaptureEnd = ulOrigCaptureEnd;
						break;
					}
				}
				else
				{
#ifdef DEBUG
printk("video.c: No TEMP region has been initialized yet.\n");
#endif
				}

				/* Update the current video capture window values */
				/* The only values matter are width and height which specify the
						capture resolution */
				VDev.gCaptureWindow.x						= VideoWindow.x;
				VDev.gCaptureWindow.y						= VideoWindow.y;
				VDev.gCaptureWindow.width				= VideoWindow.width;
				VDev.gCaptureWindow.height			= VideoWindow.height;
				VDev.gCaptureWindow.chromakey		= VideoWindow.chromakey;
				VDev.gCaptureWindow.flags				= VideoWindow.flags;
				VDev.gCaptureWindow.clips				= VideoWindow.clips;
				VDev.gCaptureWindow.clipcount		= VideoWindow.clipcount;

				VDev.gnSubCaptureOn							= 0;			/* sub-capture diactivated */

				/*
					Actual capture width/height will be modified if subcapture
					takes place.
				*/
				VDev.gnActualCaptureWidth		= VDev.gCaptureWindow.width;
				VDev.gnActualCaptureHeight	= VDev.gCaptureWindow.height;

				/* Restore the original capture clipping settings */
				if (NTSC == VDev.gnVideoStandard)
				{
					Out_Video_Reg(CAPTURE_H_START_L, 0x0F);	
					Out_Video_Reg(CAPTURE_H_START_H, 0x00); 
					Out_Video_Reg(CAPTURE_H_END_L,   0x90);  
					Out_Video_Reg(CAPTURE_H_END_H,   0x05);

					Out_Video_Reg(CAPTURE_V_START_L, 0x0A);
					Out_Video_Reg(CAPTURE_V_START_H, 0x00);
					Out_Video_Reg(CAPTURE_V_END_L,   0xFD);
					Out_Video_Reg(CAPTURE_V_END_H,   0x00);
				}
				else if (PAL == VDev.gnVideoStandard)
				{
					Out_Video_Reg(CAPTURE_H_START_L, 0x13);
					Out_Video_Reg(CAPTURE_H_START_H, 0x00);
					Out_Video_Reg(CAPTURE_H_END_L,   0x93);
					Out_Video_Reg(CAPTURE_H_END_H,   0x05);

					Out_Video_Reg(CAPTURE_V_START_L, 0x0B);
					Out_Video_Reg(CAPTURE_V_START_H, 0x00);
					Out_Video_Reg(CAPTURE_V_END_L,   0x2F);
					Out_Video_Reg(CAPTURE_V_END_H,   0x01);
				}

				/* The following codes are needed when capture height is < 45.
					 If the following codes are omitted, a garbage line will be seen
					 when capturing height is < 45 lines. */
				if (VDev.gCaptureWindow.height < 45)
				{
       		Out_Video_Reg(CAPTURE_V_START_L, 0x09);
				}
				else
				{
       		Out_Video_Reg(CAPTURE_V_START_L, 0x0A);
				}

				/* This variable will be used in display codes */
				VDev.gwWidthBytes = VDev.gnActualCaptureWidth * 2;

				/* Setting the start of capture address */
				IGS_SetCaptureAddr();

				/* Pass out the updated inputed values */
				if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_window*)arg,
																	sizeof(VideoWindow)))
				{
					printk("video.c: VIDIOCSWIN caused -EFAULT on verify_area.\n");
					nReturnValue = -EFAULT;
					break;
				}
				memcpy_tofs((struct video_window*)arg, &VideoWindow, 
											sizeof(VideoWindow));				
			}/* end-else */
		}
		break;

		case VIDIOCSCGWIN:		/* Getting sub-capture window parameters */
		{
			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_WRITE, 
																	(struct video_capture*)arg, 
																	sizeof(VDev.gSubCaptureWindow)))
      {
        printk("video.c: VIDIOCSCGWIN caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else 
			{
				memcpy_tofs((struct video_capture*)arg, 
										&(VDev.gSubCaptureWindow), 
										sizeof(VDev.gSubCaptureWindow));
			}
		}
		break;

		case VIDIOCSCSWIN:		/* Setting sub-capture window parameters */
		{
			struct video_capture VideoCapture;
			unsigned short usNewCaptureHStart	= 0;
			unsigned short usNewCaptureHEnd		= 0;
			unsigned short usNewCaptureVStart	= 0;
			unsigned short usNewCaptureVEnd		= 0;

			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_READ, (struct video_capture*)arg,
																	sizeof(VideoCapture)))
			{
				printk("video.c: VIDIOCSCSWIN caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_fromfs(&VideoCapture, (struct video_capture*)arg,
										sizeof(VideoCapture));

			/* Validating the input values */
			VideoCapture.x 				= (VideoCapture.x / 2) * 2;
			VideoCapture.y				= (VideoCapture.y / 2) * 2;
			VideoCapture.width		= (VideoCapture.width  / 4) * 4;
			VideoCapture.height		= (VideoCapture.height / 2) * 2;
			VideoCapture.decimation		= 0;		/* Not used */	
			VideoCapture.flags				= 0;		/* Not used */ 

			if (VideoCapture.x < 0)
			{
				VideoCapture.x = 0;
			}
			if (VideoCapture.y < 0)
			{
				VideoCapture.y = 0;
			}
			if (VideoCapture.width < MIN_CLIP_REGION_WIDTH)
			{
				VideoCapture.width = MIN_CLIP_REGION_WIDTH;
			}
			if (VideoCapture.height < MIN_CLIP_REGION_HEIGHT)
			{
				VideoCapture.height = MIN_CLIP_REGION_HEIGHT;
			}

			if ((VDev.gCaptureWindow.width - VideoCapture.x) < MIN_CLIP_REGION_WIDTH)
			{
				VideoCapture.x = VDev.gCaptureWindow.width - MIN_CLIP_REGION_WIDTH;
			}
			if ((VDev.gCaptureWindow.height - VideoCapture.y) < MIN_CLIP_REGION_HEIGHT)
			{
				VideoCapture.y = VDev.gCaptureWindow.height - MIN_CLIP_REGION_HEIGHT;
			}

			if (VideoCapture.x + VideoCapture.width > VDev.gCaptureWindow.width)
			{
				VideoCapture.width = VDev.gCaptureWindow.width - VideoCapture.x;
				VideoCapture.width = (VideoCapture.width / 4) * 4;
				VideoCapture.x = VDev.gCaptureWindow.width - VideoCapture.width;
			}
			if (VideoCapture.y + VideoCapture.height > VDev.gCaptureWindow.height)
			{
				VideoCapture.height = VDev.gCaptureWindow.height - VideoCapture.y;
			}

			/* Updating the global variables */
			VDev.gSubCaptureWindow.x						= VideoCapture.x;
			VDev.gSubCaptureWindow.y						= VideoCapture.y;
			VDev.gSubCaptureWindow.width				= VideoCapture.width;	
			VDev.gSubCaptureWindow.height				= VideoCapture.height;
			VDev.gSubCaptureWindow.decimation		= VideoCapture.decimation;
			VDev.gSubCaptureWindow.flags				= VideoCapture.flags;

			VDev.gnSubCaptureOn					= 1;		/* sub-capture activated */

			/* Setting new subcapture area */
			usNewCaptureHStart	= 0x0F + 
														(unsigned short)((VDev.gSubCaptureWindow.x * 
														2 * VDev.gVideoCapability.maxwidth) / 
														VDev.gCaptureWindow.width);
			if (usNewCaptureHStart == ((usNewCaptureHStart / 2) * 2))
			{
				usNewCaptureHStart -= 1;
			}

			usNewCaptureHEnd		=	0x0F + (unsigned short)
														(((VDev.gSubCaptureWindow.x + 
														VDev.gSubCaptureWindow.width) * 
														2 * VDev.gVideoCapability.maxwidth) / 
														VDev.gCaptureWindow.width);

			if (VDev.gCaptureWindow.height <= 112)
			{
				if ((VDev.gCaptureWindow.height <= 44) &&
						(VDev.gCaptureWindow.height == 
						(VDev.gSubCaptureWindow.y + VDev.gSubCaptureWindow.height)))
				{
					usNewCaptureVStart	= 0x09 + (unsigned short)
																((VDev.gSubCaptureWindow.y * 240) /
																VDev.gCaptureWindow.height);
					usNewCaptureVEnd		= 0x09 + (unsigned short)
																(((VDev.gSubCaptureWindow.y + 
																VDev.gSubCaptureWindow.height) *
																240) / VDev.gCaptureWindow.height) + 4;
				}
				else
				{
					usNewCaptureVStart	= 0x0A + (unsigned short)
																((VDev.gSubCaptureWindow.y * 240) /
																VDev.gCaptureWindow.height);
					usNewCaptureVEnd		= 0x0A + (unsigned short)
																(((VDev.gSubCaptureWindow.y + 
																VDev.gSubCaptureWindow.height + 1) *
																240) / VDev.gCaptureWindow.height) + 1;
				}
			}
			else if ((VDev.gCaptureWindow.height <= 240) && 
							(VDev.gCaptureWindow.height > 112))
			{
				usNewCaptureVStart	= 0x0A + (unsigned short)
															((VDev.gSubCaptureWindow.y * 240) /
															VDev.gCaptureWindow.height);
				usNewCaptureVEnd		= 0x0A + (unsigned short)
															(((VDev.gSubCaptureWindow.y + 
															VDev.gSubCaptureWindow.height + 1) *
															240) / VDev.gCaptureWindow.height) + 1;
			}
			else
			{
				usNewCaptureVStart	= 0x0A + (unsigned short)
															((VDev.gSubCaptureWindow.y * 240) /
															VDev.gCaptureWindow.height);
				usNewCaptureVEnd		= 0x0A + (unsigned short)
															(((VDev.gSubCaptureWindow.y + 
															VDev.gSubCaptureWindow.height) *
															240) / VDev.gCaptureWindow.height) + 4;
			}

			Out_Video_Reg(CAPTURE_H_START_L, LOBYTE(usNewCaptureHStart));	
			Out_Video_Reg(CAPTURE_H_START_H, HIBYTE(usNewCaptureHStart)); 
			Out_Video_Reg(CAPTURE_H_END_L,   LOBYTE(usNewCaptureHEnd));
			Out_Video_Reg(CAPTURE_H_END_H,   HIBYTE(usNewCaptureHEnd));

			Out_Video_Reg(CAPTURE_V_START_L, LOBYTE(usNewCaptureVStart));
			Out_Video_Reg(CAPTURE_V_START_H, HIBYTE(usNewCaptureVStart));
			Out_Video_Reg(CAPTURE_V_END_L,   LOBYTE(usNewCaptureVEnd));
			Out_Video_Reg(CAPTURE_V_END_H,   HIBYTE(usNewCaptureVEnd));

			/* Update the "actual" image data width and height */
			VDev.gnActualCaptureWidth		= VDev.gSubCaptureWindow.width;
			VDev.gnActualCaptureHeight	= VDev.gSubCaptureWindow.height;

			/* This variable will be used in display codes */
			VDev.gwWidthBytes = VDev.gnActualCaptureWidth * 2;

			IGS_SetCaptureAddr();

			/* Update inputed values */
			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_capture*)arg,
																	sizeof(VideoCapture)))
			{
				printk("video.c: VIDIOCSCSWIN caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_tofs((struct video_capture*)arg, &VideoCapture,
										sizeof(VideoCapture));
		}
		break;

		case VIDIOCCAPTURE:		/* Start/stop capturing */
		{
			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (UNINITIALIZED != VDev.gnSelectedChannel)
			{
				if (1 == (int)arg)
				{
					IGS_VideoCaptureOn();
				}
				else if (0 == (int)arg)
				{
					IGS_VideoCaptureOff();
				}
				else
				{
					/* Inappropriate argument, should be either 0 or 1 */
					nReturnValue = -1;
				}
			}
			else
			{
				/* No channel has been selected yet, cannot start capturing */
#ifdef DEBUG
printk("video.c: No valid channel has been selected yet.\n");
#endif
				nReturnValue = -1;
			}
		}
		break;

		case VIDIOCGCHAN:		/* Get properties of a specific channel */
		{
			struct video_channel VideoChannel;

			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_READ, (struct video_channel*)arg,
																	sizeof(VideoChannel)))
			{
				printk("video.c: VIDIOCGCHAN caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else
			{
				memcpy_fromfs(&VideoChannel, (struct video_channel*)arg,
											sizeof(VideoChannel));				

				if ((VideoChannel.channel > 0) && 
						(VideoChannel.channel <= TOTAL_NUMBER_OF_CHANNELS))
				{
					if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_channel*)arg,
																			sizeof(VideoChannel)))
      		{
        		printk("video.c: VIDIOCGCHAN caused -EFAULT on verify_area.\n");
						nReturnValue = -EFAULT;
					}
					else 
					{
						memcpy_tofs((struct video_channel*)arg, 
												&(VDev.gVideoChannel[VideoChannel.channel-1]),
												sizeof(VideoChannel));
					}
				}
				else
				{		
					/* 
						The only channel available is channel 1 - the video camera
					 	Anything else results error
					*/
					nReturnValue = -1;
				}/* end-else */
			}/* end-else */
		}
		break;

		case VIDIOCSCHAN:		/* To select a particular channel as video input */
		{
			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if ( ((int)arg > 0) && ((int)arg <= TOTAL_NUMBER_OF_CHANNELS) )
			{
				VDev.gnSelectedChannel = (int)arg;
			}
			else
			{
				/* No other channels are available */
				VDev.gnSelectedChannel = UNINITIALIZED;
				nReturnValue = -1;
			}
		}
		break;

		case VIDIOCGPICT:		/* Getting image properties */
		{
			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_picture*)arg,
																	sizeof(VDev.gVideoPicture)))
     	{
     		printk("video.c: VIDIOCGPICT caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else 
			{
				memcpy_tofs((struct video_picture*)arg, &(VDev.gVideoPicture),
										sizeof(VDev.gVideoPicture));
			}
		}
		break;

		/* Set image properties, whiteness and depth are ignored */
		case VIDIOCSPICT:	
		{
			struct video_picture	VideoPicture;
			unsigned char					uchNewValue = 0;
			int										lnTempValue = 0L;
			
			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_READ, (struct video_picture*)arg,
																	sizeof(VideoPicture)))
			{
				printk("video.c: VIDIOCSPICT caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_fromfs(&VideoPicture, (struct video_picture*)arg,
										sizeof(VideoPicture));

			/* Only update the value(s) that has been changed. */
			if (VDev.gVideoPicture.brightness != VideoPicture.brightness)
			{
				/* Scale down from 0-65535 to 0-255 (hardware brightness range) */
				if (VideoPicture.brightness + 1 <= 65535)
				{
					VideoPicture.brightness++;
				}
				VDev.gVideoPicture.brightness = VideoPicture.brightness;
				lnTempValue = VDev.gVideoPicture.brightness * 255 / 65535;
				uchNewValue = (unsigned char)lnTempValue;

				/* Setting brightness register */
				sendOneByte(P7111Port, 0x0A, uchNewValue);
			}

			if (VDev.gVideoPicture.contrast != VideoPicture.contrast)
			{
				/* Scale down from 0-65535 to 0-127 (hardware contrast range) */
				if (VideoPicture.contrast + 1 <= 65535)
				{
					VideoPicture.contrast++;
				}
				VDev.gVideoPicture.contrast = VideoPicture.contrast;
				lnTempValue = VDev.gVideoPicture.contrast * 127 / 65535;
				uchNewValue = (unsigned char)lnTempValue;

				/* Setting contrast register */
				sendOneByte(P7111Port, 0x0B, uchNewValue);
			}

			if (VDev.gVideoPicture.colour != VideoPicture.colour)
			{
				/* Scale down from 0-65535 to 0-127 (hardware saturation(colour) 
						range) */
				if (VideoPicture.colour + 1 <= 65535)
				{
					VideoPicture.colour++;
				}
				VDev.gVideoPicture.colour = VideoPicture.colour;
				lnTempValue = VDev.gVideoPicture.colour * 127 / 65535;
				uchNewValue = (unsigned char)lnTempValue;

				/* Setting saturation(colour) register */
				sendOneByte(P7111Port, 0x0C, uchNewValue);
			}

			if (VDev.gVideoPicture.hue != VideoPicture.hue)
			{
				/* Scale down from 0-65535 to 0-128 (hardware hue range) */
				if (VideoPicture.hue + 1 <= 65535)
				{
					VideoPicture.hue++;
				}
				VDev.gVideoPicture.hue = VideoPicture.hue;
				lnTempValue = VDev.gVideoPicture.hue * 128 / 65535;
				uchNewValue = (unsigned char)lnTempValue;

				/* Setting hue register */
				sendOneByte(P7111Port, 0x0D, uchNewValue);
			}

			/* 
				Setting Video palette is same as specifying the video format will be
				used in capturing.
				But right now, only YUYV (CCIR656-8bits) can be supported by hardware.
				RGB565 and RGB24 can be supported through software format converstion.
				All other capture formats are not supported at this moment.
			*/
			if (VIDEO_PALETTE_RGB565 == VideoPicture.palette) 
			{
				VDev.gVideoPicture.palette	= VideoPicture.palette;
				VDev.gVideoPicture.depth		= 16;
			}
			else if (VIDEO_PALETTE_RGB24 == VideoPicture.palette)
			{
				VDev.gVideoPicture.palette	= VideoPicture.palette;
				VDev.gVideoPicture.depth		= 24;
			}
			else	/* anything else will result the use of YUYV capture format */ 
			{
				VideoPicture.palette				= VIDEO_PALETTE_YUYV;
				VDev.gVideoPicture.palette	= VideoPicture.palette;
				VDev.gVideoPicture.depth		= 16;
			}

			/* Update inputed values */
			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_picture*)arg,
																	sizeof(VideoPicture)))
			{
				printk("video.c: VIDIOCSPICT caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_tofs((struct video_picture*)arg, &VideoPicture,
										sizeof(VideoPicture));
		}
		break;
			
		case VIDIOCGTUNER:		/* Get tuner properties */
		case VIDIOCSTUNER:		/* Set tuner properties */
		case VIDIOCGFREQ:			/* Get tuning frequency */
		case VIDIOCSFREQ:			/* Set tuning frequency */
		case VIDIOCGAUDIO:		/* Get TV/Radio audio properties */
		case VIDIOCSAUDIO:		/* Set TV/Radio audio properties */
		case VIDIOCGMBUF:
		case VIDIOCMCAPTURE:
		case VIDIOCSYNC:
		{
			/* Not supported */
			nReturnValue = -1;
		}
		break;

		case VIDIOCGUNIT:			/* Get the unit numbers of the associated devices */
		{
			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_unit*)arg,
																	sizeof(VDev.gVideoUnit)))
      {
        printk("video.c: VIDIOCGUNIT caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else 
			{
      	memcpy_tofs((struct video_unit*)arg, &(VDev.gVideoUnit), 
										sizeof(VDev.gVideoUnit));
			}
		}
		break;

		/*
		 *	A D D I T I O N A L   F U N C T I O N S
		 */
		case VIDIOC_COREL_CFREEZEONOFF:		/* Freeze/Unfreeze capturing */
		{
			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (1 == (int)arg)
			{
				VDev.gnCaptureFreeze = 1;
				IGS_FreezeVideo();
			}
			else if (0 == (int)arg)
			{
				VDev.gnCaptureFreeze = 0;
				IGS_UnfreezeVideo();
			}
			else
			{
				/* Inappropriate argument */
				nReturnValue = -1;
			}
		}
		break;
			
		case VIDIOC_COREL_MIRRORX:		/* Enable/Disable X axis mirroring */
		{
			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (1 == (int)arg)
			{
				VDev.gnCaptureXMirror = 1;

				IGS_SetXYMirror(VDev.gnCaptureXMirror, VDev.gnCaptureYMirror);
			}
			else if (0 == (int)arg)
			{
				VDev.gnCaptureXMirror = 0;

				IGS_SetXYMirror(VDev.gnCaptureXMirror, VDev.gnCaptureYMirror);
			}
			else
			{
				/* Inappropriate argument */
				nReturnValue = -1;
			}
		}
		break;

		case VIDIOC_COREL_MIRRORY:		/* Enable/Disable Y axis mirroring */
		{
			if (0 != MINOR(inodep->i_rdev))	/* Only /dev/video0 can access this */
			{
				nReturnValue = -1;
				break;
			}

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (1 == (int)arg)
			{
				VDev.gnCaptureYMirror = 1;

				IGS_SetXYMirror(VDev.gnCaptureXMirror, VDev.gnCaptureYMirror);
			}
			else if (0 == (int)arg)
			{
				VDev.gnCaptureYMirror = 0;

				IGS_SetXYMirror(VDev.gnCaptureXMirror, VDev.gnCaptureYMirror);
			}
			else
			{
				/* Inappropriate argument */
				nReturnValue = -1;
			}
		}
		break;

		case VIDIOC_COREL_GDWIN:	/* Getting display windows parameters */
		{
			int nMinor = MINOR(inodep->i_rdev);

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct display_window*)arg,
																	sizeof(VDev.gDisplayWindow[nMinor])))
      {
        printk("video.c: VIDIOC_COREL_GDWIN caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else 
			{
      	memcpy_tofs((struct display_window*)arg, 
										&(VDev.gDisplayWindow[nMinor]), 
										sizeof(VDev.gDisplayWindow[nMinor]));
			}
		}
		break;

		case VIDIOC_COREL_SDWIN:	/* Setting display windows parameters */
		{
			struct display_window	DisplayWindow;
			unsigned long					ulOrigDisplayStart	= 0L;
			unsigned long					ulOrigRegionSize		= 0L;
			int										nOrigDWinRegion			= -1;
			int										nCurrentRegion			= -1;	
			int										nDataSize						= 0;
			int										nNewFactor					= 0;
			int										nMinor							= MINOR(inodep->i_rdev);
			int										nTemp								= 0;

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_READ, (struct display_window*)arg,
																	sizeof(DisplayWindow)))
			{
				printk("video.c: VIDIOC_COREL_SDWIN caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_fromfs(&DisplayWindow, (struct display_window*)arg,
										sizeof(DisplayWindow));

			/* Validating input values (Not bullet proof) */
			if ((DisplayWindow.SrcPalette != VIDEO_PALETTE_YUYV) && 
					(DisplayWindow.SrcPalette != VIDEO_PALETTE_RGB565) &&
					(DisplayWindow.SrcPalette != VIDEO_PALETTE_RGB24))
			{
				DisplayWindow.SrcPalette = VIDEO_PALETTE_YUYV;
			}
			DisplayWindow.SrcWidth	= (DisplayWindow.SrcWidth		/ 4) * 4;
			DisplayWindow.SrcHeight = (DisplayWindow.SrcHeight	/ 2) * 2;
			if (VDev.gVideoCapability.minwidth > DisplayWindow.SrcWidth)
				DisplayWindow.SrcWidth = VDev.gVideoCapability.minwidth;
			if (VDev.gVideoCapability.minheight > DisplayWindow.SrcHeight)
				DisplayWindow.SrcHeight = VDev.gVideoCapability.minheight;
			if (DisplayWindow.dest_win.width < 0)
				DisplayWindow.dest_win.width = 0;
			if (DisplayWindow.dest_win.height < 0)
				DisplayWindow.dest_win.height = 0;
			DisplayWindow.dest_win.flags			= 0;					/* Not used */
			DisplayWindow.dest_win.clips			= (void*)0;		/* Not used */
			DisplayWindow.dest_win.clipcount	= 0;					/* Not used */

			VDev.gDisplayWindow[nMinor].SrcType				= DisplayWindow.SrcType;
			if (!(VDev.gDisplayWindow[nMinor].SrcType)) 	/* overlaying local image */
			{
				/* use local values instead of passed argument */
				DisplayWindow.SrcPalette	= VDev.gVideoPicture.palette;
				DisplayWindow.SrcWidth		= VDev.gnActualCaptureWidth;
				DisplayWindow.SrcHeight		= VDev.gnActualCaptureHeight;
				VDev.gDisplayWindow[nMinor].SrcPalette	= DisplayWindow.SrcPalette;
				VDev.gDisplayWindow[nMinor].SrcWidth		= DisplayWindow.SrcWidth;
				VDev.gDisplayWindow[nMinor].SrcHeight		= DisplayWindow.SrcHeight;
				VDev.gInputBPP[nMinor] = 2;			/* Local data always YUYV formated */
			}
			else	/* overlaying network/applications video data */
			{
				/*
					Need to reserve memory space for a particular window to overlay
					network data, expand the region if necessary or return error.
				*/
				/* use the passed arguments */
				VDev.gDisplayWindow[nMinor].SrcPalette	= DisplayWindow.SrcPalette;
				if ((nMinor != VDev.gIconBase) && (nMinor != VDev.gShiftDest))
				{
					VDev.gDisplayWindow[nMinor].SrcWidth	= DisplayWindow.SrcWidth;
					VDev.gDisplayWindow[nMinor].SrcHeight	= DisplayWindow.SrcHeight;
				}

				/* find out the mem size needed for incoming video data */
				switch(VDev.gDisplayWindow[nMinor].SrcPalette)
				{
					case VIDEO_PALETTE_YUYV:
					case VIDEO_PALETTE_RGB565:
					{
						nDataSize = VDev.gDisplayWindow[nMinor].SrcWidth *
												VDev.gDisplayWindow[nMinor].SrcHeight * 2;
					}
					break;

					case VIDEO_PALETTE_RGB24:
					{
						nDataSize = VDev.gDisplayWindow[nMinor].SrcWidth *					
												VDev.gDisplayWindow[nMinor].SrcHeight * 3;
					}
					break;
				
					default:
					{	
						nReturnValue = -1;
					}
					break;
				}

				if (-1 == nReturnValue)
				{
#ifdef DEBUG
printk("video.c: Selected overlay palette is not supported.\n");
#endif
					break;
				}
				
				/* CHECK TO SEE IF ANY DISPLAY REGION HAS BEEN ASSOCIATED WITH THE
						DISPLAY WINDOW */
				nCurrentRegion = VDev.gDWinRegion[nMinor];

				if (-1 != nCurrentRegion)
				{
#ifdef DEBUG
printk("\nvideo.c: /dev/video%d has associated to display region #%d.\n", 
				nMinor, nCurrentRegion);
#endif

					VDev.gnScaleDownFactor[nMinor] = 
																VDev.gDisplayWindow[nMinor].SrcWidth / 
																DisplayWindow.dest_win.width;
	
					if (2 <= VDev.gnScaleDownFactor[nMinor])
					{
						/* Make sure the scale down factor is in power of 2 */
						nTemp = VDev.gnScaleDownFactor[nMinor] / 2;
						nNewFactor = 2;	
						while(1 != nTemp)
						{
							nTemp /= 2;
							nNewFactor *= 2;
						}
						VDev.gnScaleDownFactor[nMinor] = nNewFactor;

						DisplayWindow.dest_win.width = 	
																		VDev.gDisplayWindow[nMinor].SrcWidth /
																		VDev.gnScaleDownFactor[nMinor];
						DisplayWindow.dest_win.height = 
																		VDev.gDisplayWindow[nMinor].SrcHeight /
																		VDev.gnScaleDownFactor[nMinor];

#ifdef DEBUG
printk("video.c: Scaling down is requested.\n");
#endif

						if (nDataSize > VDev.gAMI.gulRegionSize[nCurrentRegion])
						{
#ifdef DEBUG
printk("video.c: Data size is larger than associated display region.\n");
#endif
							ulOrigDisplayStart = 
							VDev.gAMI.gulDisplayStart[VDev.gDWinRegion[nMinor]];
							ulOrigRegionSize = 
							VDev.gAMI.gulRegionSize[VDev.gDWinRegion[nMinor]];
#ifdef DEBUG
printk("video.c: Saving all values: start: %ld size: %ld\n",
				ulOrigDisplayStart, ulOrigRegionSize);
#endif

							if (-1 != Expand_Display_Region(nMinor, nDataSize))
							{
								if (nDataSize > VDev.gAMI.gulTempSize)
								{
#ifdef DEBUG
if (VDev.gAMI.gulTempSize)
printk("video.c: Expanding existing TEMP region.\n");
else
printk("video.c: Creating TEMP region.\n");
#endif
									if (-1 == Set_Temp_Region(nDataSize))
									{
#ifdef DEBUG
printk("video.c: Restore original values.\n");
#endif
										VDev.gAMI.gulDisplayStart[VDev.gDWinRegion[nMinor]] = 
										ulOrigDisplayStart;
										VDev.gAMI.gulRegionSize[VDev.gDWinRegion[nMinor]] = 
										ulOrigRegionSize;
										nReturnValue = -1;
										break;
									}
								}
							}
							else
							{
								nReturnValue = -1;
								break;
							}
						}
						else
						{
#ifdef DEBUG
printk("video.c: No expansion is needed.\n");
#endif
							if (nDataSize > VDev.gAMI.gulTempSize)
							{
#ifdef DEBUG
if (VDev.gAMI.gulTempSize)
printk("video.c: Expanding existing TEMP region.\n");
else
printk("video.c: Creating TEMP region.\n");
#endif
								if (-1 == Set_Temp_Region(nDataSize))
								{
									nReturnValue = -1;
									break;
								}
							}
						}
					}
					else	/* No scaling down */
					{
						if (nDataSize > VDev.gAMI.gulRegionSize[nCurrentRegion])
						{
#ifdef DEBUG
printk("video.c: Data size is larger than associated display region.\n");
#endif
							if (-1 == Expand_Display_Region(nMinor, nDataSize))
							{
								nReturnValue = -1;
								break;
							}
						}
						else
						{
							/* No problem */
#ifdef DEBUG
printk("video.c: No expansion is needed.\n");
#endif
						}
					}
				}
				else	/* NO DISPLAY REGION HAS BEEN ASSOCIATED WITH THIS DISPLAY 
								WINDOW YET. */
				{
#ifdef DEBUG
printk("\nvideo.c: /dev/video%d has NOT associated to any display region.\n", 
				nMinor);
#endif

					/* Doing scaling down? */
					VDev.gnScaleDownFactor[nMinor] = 
																	VDev.gDisplayWindow[nMinor].SrcWidth /
																	DisplayWindow.dest_win.width;

					/* Initialize a display region and try to see if the necessary space
							can be reserved */
					if (2 <= VDev.gnScaleDownFactor[nMinor])	/* Scaling down */
					{
						/* Make sure scale down factor is in power of 2 */
						nTemp = VDev.gnScaleDownFactor[nMinor] / 2;
						nNewFactor = 2;	
						while(1 != nTemp)
						{
							nTemp /= 2;
							nNewFactor *= 2;
						}
						VDev.gnScaleDownFactor[nMinor] = nNewFactor;

						DisplayWindow.dest_win.width = 	
																		VDev.gDisplayWindow[nMinor].SrcWidth /
																		VDev.gnScaleDownFactor[nMinor];
						DisplayWindow.dest_win.height =
																		VDev.gDisplayWindow[nMinor].SrcHeight /
																		VDev.gnScaleDownFactor[nMinor];

#ifdef DEBUG
printk("video.c: Scaling down is requested.\n");
#endif

						ulOrigDisplayStart = 0;
						ulOrigRegionSize = 0;
						nOrigDWinRegion = -1;
#ifdef DEBUG
printk("video.c: Saving original values: start: %ld size: %ld region: %d\n",
				ulOrigDisplayStart, ulOrigRegionSize, nOrigDWinRegion);
#endif

						if (-1 != Init_Display_Region(nMinor, nDataSize))
						{
							if (nDataSize > VDev.gAMI.gulTempSize)
							{	
#ifdef DEBUG
if (VDev.gAMI.gulTempSize)
printk("video.c: Expanding existing TEMP region.\n");
else
printk("video.c: Creating TEMP region.\n");
#endif
								if (-1 == Set_Temp_Region(nDataSize))
								{
#ifdef DEBUG
printk("video.c: Restored original values.\n");
#endif
									VDev.gAMI.gulDisplayStart[VDev.gDWinRegion[nMinor]] = 
									ulOrigDisplayStart;
									VDev.gAMI.gulRegionSize[VDev.gDWinRegion[nMinor]] = 
									ulOrigRegionSize;
									VDev.gDWinRegion[nMinor] = nOrigDWinRegion;
									nReturnValue = -1;
									break;
								}
							}
						}
						else
						{
							nReturnValue = -1;
							break;
						}

					}
					else	/* No scaling down */
					{
						if (-1 == Init_Display_Region(nMinor, nDataSize))
						{
							nReturnValue = -1;	
							break;
						}
					}

				}/* end-else */
			}

			VDev.gDisplayWindow[nMinor].UseChromaKey	= DisplayWindow.UseChromaKey;
			VDev.gDisplayWindow[nMinor].dest_win.x		= DisplayWindow.dest_win.x;
			VDev.gDisplayWindow[nMinor].dest_win.y		= DisplayWindow.dest_win.y;

			if (nMinor != VDev.gIconBase)
			{
				VDev.gDisplayWindow[nMinor].dest_win.width = 
											DisplayWindow.dest_win.width;
				VDev.gDisplayWindow[nMinor].dest_win.height = 
											DisplayWindow.dest_win.height;
			}
			VDev.gDisplayWindow[nMinor].dest_win.chromakey =
											DisplayWindow.dest_win.chromakey;
			VDev.gDisplayWindow[nMinor].dest_win.flags = 
											DisplayWindow.dest_win.flags;
			VDev.gDisplayWindow[nMinor].dest_win.clips = 
											DisplayWindow.dest_win.clips;
			VDev.gDisplayWindow[nMinor].dest_win.clipcount = 
											DisplayWindow.dest_win.clipcount;

			switch(nMinor)
			{
				case 0:			/* Main video window */
				{
					/* Setting main video overlaying FORMAT (YUV422, RGB565 or RGB24) */
					if (!VDev.gDisplayWindow[0].SrcType)
					{
						/* local video must be in YUV format */
						IGS_SetCaptureFormat(YUV422);	
						VDev.gInputBPP[nMinor] = 2;
					}
					else
					{
						if (VIDEO_PALETTE_YUYV == VDev.gDisplayWindow[0].SrcPalette)
						{
							IGS_SetCaptureFormat(YUV422);
							VDev.gInputBPP[nMinor] = 2;
						}
						else if (VIDEO_PALETTE_RGB565 == VDev.gDisplayWindow[0].SrcPalette)
						{
							IGS_SetCaptureFormat(RGB565);
							VDev.gInputBPP[nMinor] = 2;
						}
						else if (VIDEO_PALETTE_RGB24 == VDev.gDisplayWindow[0].SrcPalette)
						{	
							IGS_SetCaptureFormat(RGB24);
							VDev.gInputBPP[nMinor] = 3;
						}
					}

					/* Setting main video overlaying TYPE (colorkey or windowkey) */
					if (VDev.gDisplayWindow[0].UseChromaKey)
					{
						IGS_SetOverlayFormat(COLORKEY);
						IGS_SetRGBKeyColor(VDev.gDisplayWindow[nMinor].dest_win.chromakey);
					}
					else
					{
						IGS_SetOverlayFormat(WINDOWKEY);
					}

					/* Scaling */
					IGS_SetScale(	VDev.gDisplayWindow[nMinor].SrcWidth,
												VDev.gDisplayWindow[nMinor].dest_win.width,
												VDev.gDisplayWindow[nMinor].SrcHeight,
												VDev.gDisplayWindow[nMinor].dest_win.height);

					/* Position the display window on the screen */
					IGS_AdjustVideoWindow(VDev.gDisplayWindow[nMinor].dest_win.x, 
																VDev.gDisplayWindow[nMinor].dest_win.y,
																VDev.gDisplayWindow[nMinor].dest_win.width,		
																VDev.gDisplayWindow[nMinor].dest_win.height);
				}
				break;
				
				case 1:			/* V2 video window */
				{
					/* Setting V2 video overlaying FORMAT (YUV422, RGB565 or RGB24) */
					if (!VDev.gDisplayWindow[1].SrcType)
					{
						/* local video must be in YUV format*/
						IGS_SetCaptureFormat_V2(YUV422);
						VDev.gInputBPP[nMinor] = 2;
					}
					else
					{
						if (VIDEO_PALETTE_YUYV == VDev.gDisplayWindow[1].SrcPalette)
						{
							IGS_SetCaptureFormat_V2(YUV422);
							VDev.gInputBPP[nMinor] = 2;
						}
						else if (VIDEO_PALETTE_RGB565 == VDev.gDisplayWindow[1].SrcPalette)
						{
							IGS_SetCaptureFormat_V2(RGB565);
							VDev.gInputBPP[nMinor] = 2;
						}
						else if (VIDEO_PALETTE_RGB24 == VDev.gDisplayWindow[1].SrcPalette)
						{	
							IGS_SetCaptureFormat_V2(RGB24);
							VDev.gInputBPP[nMinor] = 3;
						}
					}

					/* Setting V2 video overlaying TYPE (colorkey or windowkey) */
					if (VDev.gDisplayWindow[1].UseChromaKey)
					{
						IGS_SetOverlayFormat_V2(COLORKEY);
						IGS_SetRGBKeyColor(VDev.gDisplayWindow[1].dest_win.chromakey);
					}
					else
					{
						IGS_SetOverlayFormat_V2(WINDOWKEY);
					}

					/* Scaling up */
					IGS_SetScale_V2(VDev.gDisplayWindow[nMinor].SrcWidth,
													VDev.gDisplayWindow[nMinor].dest_win.width,
													VDev.gDisplayWindow[nMinor].SrcHeight,
													VDev.gDisplayWindow[nMinor].dest_win.height);

					/* Position the display window on the screen */
					IGS_AdjustVideoWindow_V2(VDev.gDisplayWindow[nMinor].dest_win.x, 
																	 VDev.gDisplayWindow[nMinor].dest_win.y,
																	 VDev.gDisplayWindow[nMinor].dest_win.width,
																	 VDev.gDisplayWindow[nMinor].dest_win.height);
				}
				break;

				case 2:			/* X2 video window */
				{
					/* Setting X2 video overlaying FORMAT (YUV422, RGB565 or RGB24) */
					if (!VDev.gDisplayWindow[2].SrcType)
					{
						/* local video must be in YUV format*/
						IGS_SetCaptureFormat_X2(YUV422);
						VDev.gInputBPP[nMinor] = 2;
					}
					else
					{
						if (VIDEO_PALETTE_YUYV == VDev.gDisplayWindow[2].SrcPalette)
						{
							IGS_SetCaptureFormat_X2(YUV422);
							VDev.gInputBPP[nMinor] = 2;
						}
						else if (VIDEO_PALETTE_RGB565 == VDev.gDisplayWindow[2].SrcPalette)
						{
							IGS_SetCaptureFormat_X2(RGB565);
							VDev.gInputBPP[nMinor] = 2;
						}
						else if (VIDEO_PALETTE_RGB24 == VDev.gDisplayWindow[2].SrcPalette)
						{	
							IGS_SetCaptureFormat_X2(RGB24);
							VDev.gInputBPP[nMinor] = 3;
						}
					}

					/* Setting X2 video overlaying TYPE (colorkey or windowkey) */
					if (VDev.gDisplayWindow[2].UseChromaKey)
					{
						IGS_SetOverlayFormat_X2(COLORKEY);
						IGS_SetRGBKeyColor(VDev.gDisplayWindow[2].dest_win.chromakey);
					}
					else
					{
						IGS_SetOverlayFormat_X2(WINDOWKEY);
					}

					/* Scaling up */
					IGS_SetScale_X2(VDev.gDisplayWindow[nMinor].SrcWidth,
													VDev.gDisplayWindow[nMinor].dest_win.width,
													VDev.gDisplayWindow[nMinor].SrcHeight,
													VDev.gDisplayWindow[nMinor].dest_win.height);

					/* Position the display window on the screen */
					IGS_AdjustVideoWindow_X2(VDev.gDisplayWindow[nMinor].dest_win.x, 
																	 VDev.gDisplayWindow[nMinor].dest_win.y,
																	 VDev.gDisplayWindow[nMinor].dest_win.width,
																	 VDev.gDisplayWindow[nMinor].dest_win.height);
				}
				break;

				default:
					printk("video.c: Should not reach here!\n");
				break;
			}/* end-switch */

			/* Update inputed values */
			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct display_window*)arg,
																	sizeof(DisplayWindow)))
			{
				printk("video.c: VIDIOC_COREL_SDWIN caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_tofs((struct display_window*)arg, &DisplayWindow,
										sizeof(DisplayWindow));
		}
		break;

		case VIDIOC_COREL_PREVIEW:	/* Start/Stop video windows' overlaying */
		{		
			/* This ioctl can handle enable/disable of all three overlay windows */
			int nMinor = MINOR(inodep->i_rdev);

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			switch(nMinor)
			{
				case 0:	/* Main video window */
				{
					if (1 == (int)arg)
					{
						IGS_VideoOn();
						VDev.gWindowOnOff[0] = 1;
					}
					else if (0 == (int)arg)
					{
						IGS_VideoOff();
						VDev.gWindowOnOff[0] = 0;
					}
					else	
					{
						printk("video.c: Invalid argument.\n");
						nReturnValue = -1;
					}
				}
				break;
	
				case 1:	/* V2 video window */
				{
					if (1 == (int)arg)
					{
						IGS_OverlayOn_V2();
						VDev.gWindowOnOff[1] = 1;
					}
					else if (0 == (int)arg)
					{
						IGS_OverlayOff_V2();
						VDev.gWindowOnOff[1] = 0;
					}
					else	
					{
						printk("video.c: Invalid argument.\n");
						nReturnValue = -1;
					}
				}
				break;

				case 2:	/* X2 video window */
				{
					if (1 == (int)arg)
					{
						IGS_OverlayOn_X2();
						VDev.gWindowOnOff[2] = 1;
					}
					else if (0 == (int)arg)
					{
						IGS_OverlayOff_X2();
						VDev.gWindowOnOff[2] = 0;
					}
					else	
					{
						printk("video.c: Invalid argument.\n");
						nReturnValue = -1;
					}
				}
				break;

				default:
				{
					printk("video.c: Error, minor number is not valid.\n");	
					nReturnValue = -1;
				}
				break;
			}/* end-switch */
		}
		break;

		case VIDIOC_COREL_GPIP:		/* get window setting of PIP */
		{
			int nMinor = MINOR(inodep->i_rdev);

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_pip*)arg,
																	sizeof(VDev.gPIP[nMinor])))
      {
        printk("video.c: VIDIOC_COREL_GPIP caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else 
			{
      	memcpy_tofs((struct video_pip*)arg, &(VDev.gPIP[nMinor]), 
										sizeof(VDev.gPIP[nMinor]));
			}
		}
		break;

		case VIDIOC_COREL_SPIP:		/* set up PIP window */
		{
			struct video_pip VideoPip;
			int nSubtract = 0;
			int nMinor = MINOR(inodep->i_rdev);

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_READ, (struct video_pip*)arg,
																	sizeof(VideoPip)))
			{
				printk("video.c: VIDIOC_COREL_SPIP caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_fromfs(&VideoPip, (struct video_pip*)arg,
										sizeof(VideoPip));

			/* Validate the inputed values */
			VideoPip.x = (VideoPip.x / 2) * 2;
			if (VideoPip.x < 0)
				VideoPip.x = 0;
			if (VideoPip.y <= 0)
				VideoPip.y = 1;
			if (VideoPip.width < 22)
				VideoPip.width = 22;
			VideoPip.width = (VideoPip.width / 2) * 2;
			if (VideoPip.width > VDev.gnActualCaptureWidth)
				VideoPip.width = VDev.gnActualCaptureWidth;
			if (VideoPip.height < 15)
				VideoPip.height = 15;
			VideoPip.height = (VideoPip.height / 2) * 2;
			if (VideoPip.height > VDev.gnActualCaptureHeight)
				VideoPip.height = VDev.gnActualCaptureHeight;
			if (VideoPip.x + VideoPip.width > VDev.gnActualCaptureWidth)
			{
				VideoPip.x = VDev.gnActualCaptureWidth - VideoPip.width;
				VideoPip.x = (VideoPip.x / 2) * 2;
			}
			if (VideoPip.y + VideoPip.height > VDev.gnActualCaptureHeight)
			{
				VideoPip.y = VDev.gnActualCaptureHeight - VideoPip.height;
			}
	
			/* Updating global structure */
			VDev.gPIP[nMinor].x				= VideoPip.x;
			VDev.gPIP[nMinor].y				= VideoPip.y;
			VDev.gPIP[nMinor].width		= VideoPip.width;
			VDev.gPIP[nMinor].height	= VideoPip.height;
			VDev.gPIP[nMinor].type		= VideoPip.type;
			VDev.gPIP[nMinor].order		= VideoPip.order;
		
			/* If background for the first kind of PIP */
			if (!VideoPip.type && !VideoPip.order)			/* <-- new added */	
			{
				if (3 == VDev.gScreen.bytes_per_pixel)
					nSubtract = 1;

				IGS_SetPIPWindow(	VDev.gPIP[nMinor].x-2, 
													VDev.gPIP[nMinor].y-1, 
													VDev.gPIP[nMinor].x + VDev.gPIP[nMinor].width 
													- nSubtract, 
													VDev.gPIP[nMinor].y + VDev.gPIP[nMinor].height, 
													VDev.gnActualCaptureWidth, 
													VDev.gnActualCaptureHeight);
			}

			/* Update inputed values */
			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_pip*)arg,
																	sizeof(VideoPip)))
			{
				printk("video.c: VIDIOC_COREL_SPIP caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_tofs((struct video_pip*)arg, &VideoPip, sizeof(VideoPip));
		}
		break;

		/* This ioctl must be called after VIDIOC_COREL_SDWIN */
		case VIDIOC_COREL_PIPONOFF:
		{
			int nMinor = MINOR(inodep->i_rdev);
			int nDataSize = 0;

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (1 == (int)arg)
			{
				if (!VDev.gPIP[nMinor].type)	/* If first type of PIP */
				{
					if (!VDev.gPIP[nMinor].order)	/* If background */  /* <- new added */
						IGS_EnablePIP(ON);
				}
				else	/* Second kind of PIP */
				{
					/* Set TEMP region if one does not exist or not big enough. */
					nDataSize = VDev.gDisplayWindow[nMinor].SrcWidth *
											VDev.gDisplayWindow[nMinor].SrcHeight *
											VDev.gInputBPP[nMinor];
					if (nDataSize > VDev.gAMI.gulTempSize)
					{
#ifdef DEBUG
printk("video.c: Setting new TEMP region required by PIP2.\n");
#endif
						if (-1 == Set_Temp_Region(nDataSize))
						{
#ifdef DEBUG
printk("video.c: Cannot set new TEMP region.  Cannot start PIP2.\n");
#endif
							nReturnValue = -1;
							break;
						}
					}
				}
				VDev.gnPIPFlag[nMinor] = 1;
			}
			else if (0 == (int)arg)
			{
				if (!VDev.gPIP[nMinor].type)
				{
					IGS_EnablePIP(OFF);
				}
				VDev.gnPIPFlag[nMinor] = 0;
			}
			else	
			{
				printk("video.c: Invalid argument.\n");
				nReturnValue = -1;
				break;
			}
		}
		break;

		case VIDIOC_COREL_GICON:
		{
			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_icon*)arg,
																	sizeof(VDev.gIcon)))
      {
        printk("video.c: VIDIOC_COREL_GICON caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else 
			{
      	memcpy_tofs((struct video_icon*)arg, &(VDev.gIcon), 
										sizeof(VDev.gIcon));
			}
		}
		break;

		/* This ioctl should be called before setting window parameters for the
				destination window. */
		case VIDIOC_COREL_SICON:
		{
			struct video_icon VideoIcon;
			int nMinor = MINOR(inodep->i_rdev);

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_READ, (struct video_icon*)arg,
																	sizeof(VideoIcon)))
			{
				printk("video.c: VIDIOC_COREL_SICON caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_fromfs(&VideoIcon, (struct video_icon*)arg,
										sizeof(VideoIcon));

			/* Validating input values */
			/* Icon width need to set to multiple of 8 to avoid problem doing RGB24
					under 24bpp */
			VideoIcon.width = (VideoIcon.width / 8) * 8;

			if (VideoIcon.width > 88)
			{
				VideoIcon.width = 88;
			}
			if (VideoIcon.width < 40)
			{
				VideoIcon.width = 40;
			}
			if (VideoIcon.height > 72)
			{
				VideoIcon.height = 72;
			}
			if (VideoIcon.height < 32)
			{
				VideoIcon.height = 32;
			}
			if ((VideoIcon.total_num < 1) || (VideoIcon.total_num > 6))
			{	
				nReturnValue = -1;
				break;
			}
			else
			{
				/* Updating global structure */
				VDev.gIcon.total_num	= VideoIcon.total_num;
				VDev.gIcon.width			= VideoIcon.width;
				VDev.gIcon.height			= VideoIcon.height;
				VDev.gIcon.src				= VideoIcon.src;

				VDev.gDisplayWindow[nMinor].SrcWidth = VDev.gIcon.width;
				VDev.gDisplayWindow[nMinor].SrcHeight = VDev.gIcon.total_num * 
																								VDev.gIcon.height ; 
																								//+2 * (VDev.gIcon.total_num - 1);
				VDev.gDisplayWindow[nMinor].dest_win.width = 
																				VDev.gDisplayWindow[nMinor].SrcWidth;
				VDev.gDisplayWindow[nMinor].dest_win.height =
																				VDev.gDisplayWindow[nMinor].SrcHeight;


				/* minor number calls this ioctl becomes the display video window */
				VDev.gIconBase = nMinor;
			}

			/* Update inputed values */
			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_icon*)arg,
																	sizeof(VideoIcon)))
			{
				printk("video.c: VIDIOC_COREL_SICON caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_tofs((struct video_icon*)arg, &VideoIcon, sizeof(VideoIcon));
		}
		break;

		case VIDIOC_COREL_ICONSELECT:
		{
			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (((int)arg >= 1) && ((int)arg <= 6))
			{
				VDev.gIconCurrent = (int)arg;
			}
			else	
			{
				printk("video.c: Invalid argument.\n");
				nReturnValue = -1;
			}
		}
		break;

		case VIDIOC_COREL_ICONONOFF:
		{
			int nMinor = MINOR(inodep->i_rdev);
			
			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if ((-1 != VDev.gIconBase) && (nMinor == VDev.gIconBase))
			{
				if (1 == (int)arg)
				{
					VDev.gIconFlag = 1;
				}
				else if (0 == (int)arg)
				{
					VDev.gIconFlag = 0;
					VDev.gIconBase = -1;
				}
				else
				{
					printk("video.c: Invalid argument.\n");
					nReturnValue = -1;
				}
			}
			else
			{
				printk("video.c: Icon parameters have not been set yet.\n");
			}
		}
		break;

		case VIDIOC_COREL_GDFREEZE:
		{
			int nMinor = MINOR(inodep->i_rdev);

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_freeze*)arg,
																	sizeof(VDev.gVideoFreeze[nMinor])))
      {
        printk("video.c: VIDIOC_COREL_GDFREEZE caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else 
			{
      	memcpy_tofs((struct video_freeze*)arg, &(VDev.gVideoFreeze[nMinor]), 
										sizeof(VDev.gVideoFreeze[nMinor]));
			}
		}
		break;

		case VIDIOC_COREL_SDFREEZE:
		{
			struct video_freeze VideoFreeze;
			int nMinor = MINOR(inodep->i_rdev);

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_READ, (struct video_freeze*)arg,
																	sizeof(VideoFreeze)))
			{
				printk("video.c: VIDIOC_COREL_SDFREEZE caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_fromfs(&VideoFreeze, (struct video_freeze*)arg,
										sizeof(VideoFreeze));

			/* Validating input values */
			VideoFreeze.x = (VideoFreeze.x / 2) * 2;
			if (VideoFreeze.x < 0)
				VideoFreeze.x = 0;
			if (VideoFreeze.y < 0)
				VideoFreeze.y = 0;

			if (VideoFreeze.width < 2)
				VideoFreeze.width = 2;
			if (VideoFreeze.width > VDev.gDisplayWindow[nMinor].SrcWidth)
				VideoFreeze.width = VDev.gDisplayWindow[nMinor].SrcWidth;
			VideoFreeze.width = (VideoFreeze.width / 2) * 2;

			if (VideoFreeze.height < 2)
				VideoFreeze.height = 2;
			if (VideoFreeze.height > VDev.gDisplayWindow[nMinor].SrcHeight)
				VideoFreeze.height = VDev.gDisplayWindow[nMinor].SrcHeight;
			VideoFreeze.height = (VideoFreeze.height / 2) * 2;

			if ((VideoFreeze.x + VideoFreeze.width) > 
					VDev.gDisplayWindow[nMinor].SrcWidth)
			{
				VideoFreeze.x = VDev.gDisplayWindow[nMinor].SrcWidth - 
														VideoFreeze.width;
				VideoFreeze.x = (VideoFreeze.x / 2) * 2;
			}
			if ((VideoFreeze.y + VideoFreeze.height) >
					VDev.gDisplayWindow[nMinor].SrcHeight)
			{
				VideoFreeze.y = VDev.gDisplayWindow[nMinor].SrcHeight -
															VideoFreeze.height;
			}

			/* Update global variables */
			VDev.gVideoFreeze[nMinor].x				= VideoFreeze.x;
			VDev.gVideoFreeze[nMinor].y				= VideoFreeze.y;
			VDev.gVideoFreeze[nMinor].width		= VideoFreeze.width;
			VDev.gVideoFreeze[nMinor].height	= VideoFreeze.height;
			VDev.gVideoFreeze[nMinor].type		= VideoFreeze.type;

			/* Update inputed values */
			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_freeze*)arg,
																	sizeof(VideoFreeze)))
			{
				printk("video.c: VIDIOC_COREL_SDFREEZE caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_tofs((struct video_freeze*)arg, &VideoFreeze,
										sizeof(VideoFreeze));
		}
		break;

		case VIDIOC_COREL_DFREEZEONOFF:
		{
			int nMinor = MINOR(inodep->i_rdev);
			int nDataSize = 0;

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (1 == (int)arg)
			{
				VDev.gFreezeFlag[nMinor] = 1;

				/* Set TEMP region if one does not exist or not big enough. */
				nDataSize = VDev.gDisplayWindow[nMinor].SrcWidth *
										VDev.gDisplayWindow[nMinor].SrcHeight *
										VDev.gInputBPP[nMinor];
				if (nDataSize > VDev.gAMI.gulTempSize)
				{
#ifdef DEBUG
printk("video.c: Setting new TEMP region required by Display Freeze.\n");
#endif
					if (-1 == Set_Temp_Region(nDataSize))
					{
#ifdef DEBUG
printk("video.c: Cannot set new TEMP region.  Cannot start PIP2.\n");
#endif
						nReturnValue = -1;
					}
				}
			}
			else if (0 == (int)arg)
			{
				VDev.gFreezeFlag[nMinor] = 0;
			}
			else
			{
				printk("video.c: Invalid argument.\n");
				VDev.gFreezeFlag[nMinor] = 0;
				nReturnValue = -1;
			}
		}
		break;

		case VIDIOC_COREL_GSHIFT:
		{
			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_shift*)arg,
																	sizeof(VDev.gVideoShift)))
			{
				printk("video.c: VIDIOC_COREL_GSHIFT caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else 
			{
      	memcpy_tofs((struct video_shift*)arg, &(VDev.gVideoShift), 
										sizeof(VDev.gVideoShift));
			}
		}
		break;

		/* whichever video window make this ioctl becomes the destination of the
				shifted video. */
		/* This ioctl should be called AFTER setting the source window parameters
				and BEFORE setting the destination window parameters */
		case VIDIOC_COREL_SSHIFT:
		{
			struct video_shift VideoShift;
			int SrcWidth = 0;
			int SrcHeight = 0;
			int nMinor = MINOR(inodep->i_rdev);

			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (-EFAULT == verify_area(	VERIFY_READ, (struct video_shift*)arg,
																	sizeof(VideoShift)))
			{
				printk("video.c: VIDIOC_COREL_SSHIFT caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_fromfs(&VideoShift, (struct video_shift*)arg,
										sizeof(VideoShift));

			/* Validating input values */
			if ((VideoShift.src < 0) || (VideoShift.src > 2))
			{
				/* Invalid source field */
				nReturnValue = -1;
				break;
			}

			/* Check if source window parameters have been set yet */
			if ((VDev.gDisplayWindow[VideoShift.src].SrcWidth <= 0) ||
					(VDev.gDisplayWindow[VideoShift.src].SrcHeight <= 0))
			{
#ifdef DEBUG
printk("video.c: Source window has not been set up yet.\n");
#endif
				nReturnValue = -1;
				break;
			}
			else
			{
				SrcWidth	= VDev.gDisplayWindow[VideoShift.src].SrcWidth;
				SrcHeight = VDev.gDisplayWindow[VideoShift.src].SrcHeight;
			}

			VideoShift.x = (VideoShift.x / 2) * 2;

			if (VideoShift.x < 0)
				VideoShift.x = 0;
			if (VideoShift.y < 0)
				VideoShift.y = 0;

			if (VideoShift.width < 8)
				VideoShift.width = 8;
			if (VideoShift.width > SrcWidth)
				VideoShift.width = SrcWidth;

			if (VideoShift.height < 8)
				VideoShift.height = 8;
			if (VideoShift.height > SrcHeight)
				VideoShift.height = SrcHeight;

			if ((VideoShift.x + VideoShift.width) > SrcWidth)
			{
				VideoShift.x = SrcWidth - VideoShift.width;
				VideoShift.x = (VideoShift.x / 2) * 2;
			}
			if ((VideoShift.y + VideoShift.height) > SrcHeight)
			{
				VideoShift.y = SrcHeight - VideoShift.height;
			}

			/* Update global variables */
			VDev.gVideoShift.x			= VideoShift.x;
			VDev.gVideoShift.y			= VideoShift.y;
			VDev.gVideoShift.width	= VideoShift.width;
			VDev.gVideoShift.height	= VideoShift.height;
			VDev.gVideoShift.src		= VideoShift.src;
			
			/* Setting the source width and height of the DESTINATION video window */
			VDev.gDisplayWindow[nMinor].SrcWidth = VDev.gVideoShift.width;
			VDev.gDisplayWindow[nMinor].SrcHeight = VDev.gVideoShift.height;

			/* This video window becomes the destination of the shifted video */
			VDev.gShiftDest = nMinor;

			/* Update inputed values */
			if (-EFAULT == verify_area(	VERIFY_WRITE, (struct video_shift*)arg,
																	sizeof(VideoShift)))
			{
				printk("video.c: VIDIOC_COREL_SSHIFT caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}

			memcpy_tofs((struct video_shift*)arg, &VideoShift, sizeof(VideoShift));
		}
		break;
	
		case VIDIOC_COREL_SHIFTONOFF:
		{
			/* Setup IGS video register, check if capture can be done and set up
					memory region for video capture/overlay. */
			if (!VDev.gnVideoInitialized && (0 != (nReturnValue = setup_video())))
			{
				printk("video.c: Error setting up video device.\n");
				break;
			}

			if (1 == (int)arg)
			{
				VDev.gShiftFlag = 1;
			}
			else if (0 == (int)arg)
			{
				VDev.gShiftDest = -1;
				VDev.gShiftFlag = 0;
			}
			else
			{
				printk("video.c: Invalid argument.\n");
				VDev.gShiftDest = -1;
				VDev.gShiftFlag = 0;
				nReturnValue = -1;
			}
		}
		break;

		case VIDIOC_COREL_S_SCREENOFFSET:
		{
			struct video_screen_params vsp;
			int i=0;

			if (-EFAULT == verify_area(	VERIFY_READ, 
																	(struct video_screen_params*)arg,
																	sizeof(vsp)))
			{
				printk("video.c: VIDIOC_COREL_S_SCREENOFFSET caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
				break;
			}
			memcpy_fromfs(&vsp, (struct video_screen_params*)arg,
										sizeof(vsp));

			if (vsp.x_ext<0 || vsp.x_ext>10000 ||
					vsp.y_ext<0 || vsp.y_ext>10000 ||
					vsp.x_orig<0 || vsp.x_orig>vsp.x_ext ||
					vsp.y_orig<0 || vsp.y_orig>vsp.y_ext)
			{	
				nReturnValue = -1;
				printk("video.c: Out of range in VIDIOC_COREL_S_SCREENOFFSET.\n");
				break;
			}

			/* Update global driver structure */
#if 0
			VDev.gScreen.x_ext = vsp.x_ext;
			VDev.gScreen.y_ext = vsp.y_ext;
#endif
			VDev.gScreen.x_orig = vsp.x_orig;
			VDev.gScreen.y_orig = vsp.y_orig;
			VDev.gScreen.width = vsp.width;
			VDev.gScreen.height = vsp.height;
			
			for (i=0; i<MAX_NUM_OF_DISPLAY_REGION; i++)
			{
				if (VDev.gWindowOnOff[i])
				{
					if (GetVideoFIFO())	
					{
						if ((VDev.gDisplayWindow[i].dest_win.width * 
								VDev.gScreen.bytes_per_pixel)
								<= VDev.gusHighResThreshold)
						{
							Change_VideoFIFO(0);
						}
						else
						{
							Change_VideoFIFO(1);
						}

						switch(i)
						{
							case 0:
								IGS_AdjustVideoWindow(VDev.gDisplayWindow[0].dest_win.x,
																			VDev.gDisplayWindow[0].dest_win.y,
																			VDev.gDisplayWindow[0].dest_win.width,
																			VDev.gDisplayWindow[0].dest_win.height);
							break;
							case 1:
								IGS_AdjustVideoWindow_V2(	VDev.gDisplayWindow[1].dest_win.x,
																					VDev.gDisplayWindow[1].dest_win.y,
																					VDev.gDisplayWindow[1].dest_win.width,
																					VDev.gDisplayWindow[1].dest_win.height);
							break;
							case 2:
								IGS_AdjustVideoWindow_X2(	VDev.gDisplayWindow[2].dest_win.x,
																					VDev.gDisplayWindow[2].dest_win.y,
																					VDev.gDisplayWindow[2].dest_win.width,
																					VDev.gDisplayWindow[2].dest_win.height);
							break;
						} /* end-switch */
					}
					else 
					{
						printk("video.c: Error getting FIFO info.\n");
					}
				}/* end-if */
			}/* end-for */

#if 0
					nNewX = max(VDev.gDisplayWindow[i].dest_win.x - 
											VDev.gScreen.x_orig, 0);
					nNewW = min(VDev.gDisplayWindow[i].dest_win.x + 
											VDev.gDisplayWindow[i].dest_win.width,
											VDev.gScreen.x_orig + VDev.gScreen.width) - 
									nNewX - VDev.gScreen.x_orig;
					nNewY = max(VDev.gDisplayWindow[i].dest_win.y - 
											VDev.gScreen.y_orig, 0);
					nNewH = min(VDev.gDisplayWindow[i].dest_win.y +
											VDev.gDisplayWindow[i].dest_win.height,
											VDev.gScreen.y_orig + VDev.gScreen.height) -
									nNewY - VDev.gScreen.y_orig;
printk("[Win%d] X: %4d Y: %4d W: %4d H: %4d\n", i, nNewX, nNewY, nNewW, nNewH);
#endif

		}
		break;

		case VIDIOC_COREL_G_SCREENOFFSET:
		{
			if (-EFAULT == verify_area(	VERIFY_WRITE, 
																	(struct video_screen_params*)arg,
																	sizeof(VDev.gScreen)))
			{
				printk("video.c: VIDIOC_COREL_G_SCREENOFFSET caused -EFAULT on verify_area.\n");
				nReturnValue = -EFAULT;
			}
			else 
			{
      	memcpy_tofs((struct video_screen_params*)arg, &(VDev.gScreen), 
										sizeof(VDev.gScreen));
			}
		}
		break;

		case VIDIOC_COREL_READONOFF:
		{
			if (1 == (int)arg)
				VDev.gnReadOn = 1;
			else if (0 == (int)arg)
				VDev.gnReadOn = 0;
			else
				nReturnValue = -1;
		}
		break;
	}

	return(nReturnValue);
}

/* When module is installed */
#ifndef MODULE
long video_init(long mem_start, long mem_emd)
{
	if (video_major = register_chrdev(0, video_device_name, &video_fops))
		printk("video.c: Unable to get major for Video Capture Device\n");
	else
		printk("video.c: video_major = %d\n", video_major);

	/* clear global structures */
	memset(&XVideoBuffer, 0, sizeof(XVideoBuffer));
	memset(&VDev, 0, sizeof(VDev));

	return mem_start;
}
#else
int init_module(void)
{
	if (register_chrdev(video_major = VIDEO_MAJOR,
											video_device_name,
											&video_fops) == -EBUSY)
	{
		printk("video.c: Unable to get major for Video Capture Device.\n");
		return -EIO;
	}
	else
		printk("video.c: video_major = %d\n", video_major);

	/* clear global structures */
	memset(&XVideoBuffer, 0, sizeof(XVideoBuffer));
	memset(&VDev, 0, sizeof(VDev));

	return 0;
}

/* When the module is removed */
void cleanup_module(void)
{
	/* Make sure all video windows are shutdown */
	IGS_OverlayOff_V2();
	IGS_VideoCleanUp_V2();

	IGS_OverlayOff_X2();
	IGS_VideoCleanUp_X2();

	IGS_VideoOff();
	IGS_VideoCleanUp();

	unregister_chrdev(video_major, video_device_name);
}
#endif
