/*
	This header file holds the global device state structure.
	This header file is used by the following files.
		igs.c
		igscop.c
		philips.c
		video_color.c
		video.c
*/

#ifndef _VIDEO_H
#define _VIDEO_H

#include "videodev.h"

/*
 *	C O N S T A N T S		(used within driver)
 */

/* Max/Min capture resolution of NTSC and PAL camera */
#define MAX_NTSC_CAPTURE_WIDTH    704
#define MAX_NTSC_CAPTURE_HEIGHT   480
#define MAX_PAL_CAPTURE_WIDTH     704       /* Temp value */
#define MAX_PAL_CAPTURE_HEIGHT    576       /* Temp value */

/* MAX and MIN resolution */
#define MIN_CAPTURE_WIDTH     		32  	/* Same for both NTSC & PAL camera */
#define MIN_CAPTURE_HEIGHT    		32  	/* Same for both NTSC & PAL camera */

/* Default capture image properties */
#define DEFAULT_CAPTURE_BRIGHTNESS      32768   /* 50% */
#define DEFAULT_CAPTURE_HUE             0       /* 0%  */
#define DEFAULT_CAPTURE_COLOUR          32768   /* 50% */
#define DEFAULT_CAPTURE_CONTRAST        36045   /* 55% */
#define DEFAULT_CAPTURE_WHITENESS				0				/* 0%  */
#define DEFAULT_CAPTURE_DEPTH						16
#define DEFAULT_CAPTURE_PALETTE					VIDEO_PALETTE_YUYV

#define		UNINITIALIZED   							-1

/*
	This indicates the number of bytes will be skipped before the first
	byte of capture data being written.
	The reason to have this is to avoid writting data into on screen
	memory region while doing X/Y mirroring.
*/
#define		BUFFER_REGION_SIZE						1408

/*
	Min values for a clipped region (Max values will depend on
	the current capture resolution)
*/
#define		MIN_CLIP_REGION_WIDTH					4
#define		MIN_CLIP_REGION_HEIGHT				2

/* Maximum number of extra lines will be captured */
#define		NTSC_EXTRA_CAP_LINES					6		/* NTSC camera (up to 704x480) */
#define		PAL_EXTRA_CAP_LINES						0		/* Temp value */

/* max no. of display region will ever be needed 
	 1. main video window
   2. V2   video window
   3. X2   video window
*/
#define		MAX_NUM_OF_DISPLAY_REGION			3
#define		TOTAL_NUMBER_OF_CHANNELS			1


/*
 *	P R I V A T E   S T R U C T U R E S
 */

/* This struct is used to communicate between X server and video driver */
struct xvideo_buffer
{
	struct video_buffer vbuf;
	int									gnWithinX:1;			/* 1 - True, 0 - false */
} XVideoBuffer;

/*
	This struct is used to manage the available memory used by capturing
	and displaying.
*/
struct avail_mem_info
{
	unsigned int	gunTotalAvailableMemory;
	unsigned long	gulAvailableMemoryStart;
	unsigned long	gulCaptureStart;
	unsigned long	gulCaptureEnd;
	unsigned long gulTempStart;
	unsigned long gulTempEnd;
	unsigned long gulTempSize;
	unsigned long	gulDisplayStart[MAX_NUM_OF_DISPLAY_REGION];
	unsigned long gulRegionSize[MAX_NUM_OF_DISPLAY_REGION];
};

/* The struct is used by VIDIOC_COREL_S/G_SCREENOFFSET */
struct video_screen_params
{
	int x_ext, y_ext;       /* horiz/vert total pixels */
	int x_orig, y_orig;     /* offset to displayed pixels */
	int width, height;      /* size of displayed pixels */
	int bytes_per_pixel;    /* 1, 2 or 3 */
	int total_video_memory;
	int display_memory_used;
	int offset_vidcap_mem;
	int offset_overlay2_mem;
};

/* The device structure, holds current state information */
struct nw_video_device
{
	/* module management */
	int 											gnOpenCount;

	/* device capability */
	struct video_capability		gVideoCapability;
	
	/* the only one video source */
	struct video_channel			gVideoChannel[TOTAL_NUMBER_OF_CHANNELS];
	int												gnSelectedChannel;

	/* Unit number of associated devices */
	struct video_unit					gVideoUnit;

	/* video tunner and video audio */
	struct video_tuner				gVideoTuner;
	struct video_audio				gVideoAudio;

	/* use by 7111 & 7110 decoder code */
	int												gnWhichDecoder;		/* which decoder are we using? */
	unsigned short						gusHighResThreshold; 	/* > 352*2 ? */

	int												gnVideoStandard;	/* NTSC or PAL */

	/* C A P T U R E   S T A T E */
	struct video_picture			gVideoPicture;			/* capture properties */
	struct video_window				gCaptureWindow;			/* normal capture window */
	struct video_capture			gSubCaptureWindow;	/* sub-capture window */
	int												gnActualCaptureWidth;		
	int												gnActualCaptureHeight;
	int												gnFullScreenOk:1;			/* 1 - true, 0 - false */
	int												gnCaptureXMirror:1;		/* 1 - ON, 0 - OFF */
	int												gnCaptureYMirror:1;		/* 1 - ON, 0 - OFF */
	int												gnCaptureFreeze:1;		/* 1 - ON, 0 - OFF */
	int												gnSubCaptureOn:1;			/* 1 - ON, 0 - OFF */
	int												gnGarbageLines;				
	int												gwWidthBytes;

	/* D I S P L A Y   S T A T E */
	struct video_screen_params	gScreen;				/* current vga settings */

	int										gWindowOnOff[3];			/* which video window has been
																									turned on 1-on, 0-off */
	struct display_window gDisplayWindow[3];		/* at most 3 overlay windows */
	int										gInputBPP[3];					/* bytes per pixel of the input
																									data */
	int										gDWinRegion[3];				/* which display region is
																									currently being used by
																									a particular display
																									window (Main, V2 or X2). 		
																									-1 - uninitialized
																									 0 - using region 0
																									 1 - using region 1
																									 2 - using region 2
																							*/
	int										gnScaleDownFactor[3];	/* Scale down factor for each
																									video window 
																									0 - main window
																									1 - V2 window
																									2 - X2 window
																							*/
	struct video_pip			gPIP[3];							/*	settings for the PIP for
																									each video window.
																							*/
	int										gnPIPFlag[3];					/*	Flag indicating whether
																									PIP is being used by a
																									particular video window.
																							*/
	struct video_icon			gIcon;								/*	settings for icon video */
	int										gIconWinWidth;				/*	same as source width of 
																									display window */
	int										gIconWinHeight;				/*	same as source height of
																									display window */
	int										gIconCurrent;					/*	Currently selected icon 
																									0 - nothing is selected
																									1 - 6 corresponding icons
																							*/
	int										gIconBase;						/*	which video window will the	
																									icon videos be shown */
	int										gIconFlag;						/*	Is Icon video being used? 
																									0 - False
																									1 - True
																							*/
	struct video_freeze		gVideoFreeze[3];			/*	settings for display freeze */
	int										gFreezeFlag[3];				/*	Is display on the particular
																									window being freezed? 
																									0 - False
																									1 - True
																							*/
	struct video_shift		gVideoShift;					/*	setting for video shift */
	int										gShiftDest;						/*	which video window will the
																									shifted video be shown */
	int										gShiftFlag:1;					/*	Is shift video being used? 
																									0 - false, 1 - true */

	/* 
		This struct is used to manage the available memory shared by capturing
		and displaying. 
	*/
	struct avail_mem_info			gAMI;

	/* Has the video device been initialized */
	int												gnVideoInitialized:1;   /* 1-true, 0-false */

	/* Indicators of whether individual window has been opened yet */
	int												gnMainOpen;	
	int												gnV2Open:1;							/* 1-true, 0-false */
	int												gnX2Open:1;							/* 1-true, 0-false */

	/* Has 7111 decoder been turned on ? */
	int												gn7111On:1;							/* 1-true, 0-false */

	/* Has video memory been assigned by X server yet? */
#if 0
	int												gnMemBufSet:1;	/* 1-true, 0-false */
#endif

	int												gnReadOn:1;							/* 1-true, 0-false */

#if SUPPORTVUNDERSCAN
	int		wV_Underscan;
	int		iMaxLines;	/* # of scan lines within which we drop one scan line */
	int		iWhichLine;	/* which scan line we drop */
#endif

} VDev;


/*
 *  U S E F U L   D E F I N I T I O N
 */
#undef  outb_p
#define outb_p(v, p) (*(volatile char*)(0xd0800000 + p) = v)
#undef  inb_p
#define inb_p(p) (*(volatile char*)(0xd0800000 + p))
#undef  outw_p
#define outw_p(v, p) outb_p(((char)(v&0xff)), (p)); \
				outb_p(((char)((v&0xff00)>>8)), (p+1));

#endif /* _VIDEO_H */
